import json
import random
import re
import time

from .common import InfoExtractor
from ..utils import (
    ExtractorError,
    clean_html,
    extract_attributes,
    float_or_none,
    format_field,
    int_or_none,
    join_nonempty,
    parse_filesize,
    parse_qs,
    str_or_none,
    strftime_or_none,
    try_get,
    unified_timestamp,
    update_url_query,
    url_or_none,
    urljoin,
)
from ..utils.traversal import find_element, find_elements, traverse_obj


class BandcampIE(InfoExtractor):
    _VALID_URL = r'https?://(?P<uploader>[^/]+)\.bandcamp\.com/track/(?P<id>[^/?#&]+)'
    _EMBED_REGEX = [r'<meta property="og:url"[^>]*?content="(?P<url>.*?bandcamp\.com.*?)"']
    _TESTS = [{
        'url': 'http://youtube-dl.bandcamp.com/track/youtube-dl-test-song',
        'md5': 'c557841d5e50261777a6585648adf439',
        'info_dict': {
            'id': '1812978515',
            'ext': 'mp3',
            'title': 'youtube-dl "\'/\\ä↭ - youtube-dl "\'/\\ä↭ - youtube-dl test song "\'/\\ä↭',
            'duration': 9.8485,
            'uploader': 'youtube-dl "\'/\\ä↭',
            'upload_date': '20121129',
            'thumbnail': r're:https?://f4\.bcbits\.com/img/.+\.jpg',
            'timestamp': 1354224127,
            'track': 'youtube-dl "\'/\\ä↭ - youtube-dl test song "\'/\\ä↭',
            'track_id': '1812978515',
            'uploader_url': 'https://youtube-dl.bandcamp.com',
            'uploader_id': 'youtube-dl',
            'artists': ['youtube-dl "\'/\\ä↭'],
            'album_artists': ['youtube-dl "\'/\\ä↭'],
        },
        'skip': 'There is a limit of 200 free downloads / month for the test song',
    }, {
        # free download
        'url': 'http://benprunty.bandcamp.com/track/lanius-battle',
        'info_dict': {
            'id': '2650410135',
            'ext': 'm4a',
            'title': 'Ben Prunty - Lanius (Battle)',
            'uploader': 'Ben Prunty',
            'thumbnail': r're:https?://f4\.bcbits\.com/img/.+\.jpg',
            'timestamp': 1396508491,
            'upload_date': '20140403',
            'release_timestamp': 1396483200,
            'release_date': '20140403',
            'duration': 260.877,
            'track': 'Lanius (Battle)',
            'track_number': 1,
            'track_id': '2650410135',
            'album': 'FTL: Advanced Edition Soundtrack',
            'uploader_url': 'https://benprunty.bandcamp.com',
            'uploader_id': 'benprunty',
            'tags': ['soundtrack', 'chiptunes', 'cinematic', 'electronic', 'video game music', 'California'],
            'artists': ['Ben Prunty'],
            'album_artists': ['Ben Prunty'],
        },
    }, {
        # no free download, mp3 128
        'url': 'https://relapsealumni.bandcamp.com/track/hail-to-fire',
        'md5': 'fec12ff55e804bb7f7ebeb77a800c8b7',
        'info_dict': {
            'id': '2584466013',
            'ext': 'mp3',
            'title': 'Mastodon - Hail to Fire',
            'uploader': 'Mastodon',
            'thumbnail': r're:https?://f4\.bcbits\.com/img/.+\.jpg',
            'timestamp': 1322005399,
            'upload_date': '20111122',
            'release_timestamp': 1076112000,
            'release_date': '20040207',
            'duration': 120.79,
            'track': 'Hail to Fire',
            'track_number': 5,
            'track_id': '2584466013',
            'album': 'Call of the Mastodon',
            'uploader_url': 'https://relapsealumni.bandcamp.com',
            'uploader_id': 'relapsealumni',
            'tags': ['Philadelphia'],
            'artists': ['Mastodon'],
            'album_artists': ['Mastodon'],
        },
    }, {
        # track from compilation album (artist/album_artist difference)
        'url': 'https://diskotopia.bandcamp.com/track/safehouse',
        'md5': '19c5337bca1428afa54129f86a2f6a69',
        'info_dict': {
            'id': '1978174799',
            'ext': 'mp3',
            'title': 'submerse - submerse - Safehouse',
            'uploader': 'submerse',
            'thumbnail': r're:https?://f4\.bcbits\.com/img/.+\.jpg',
            'timestamp': 1480779297,
            'upload_date': '20161203',
            'release_timestamp': 1481068800,
            'release_date': '20161207',
            'duration': 154.066,
            'track': 'submerse - Safehouse',
            'track_number': 3,
            'track_id': '1978174799',
            'album': 'DSK F/W 2016-2017 Free Compilation',
            'uploader_url': 'https://diskotopia.bandcamp.com',
            'uploader_id': 'diskotopia',
            'tags': ['Japan'],
            'artists': ['submerse'],
            'album_artists': ['Diskotopia'],
        },
    }]
    _WEBPAGE_TESTS = [{
        # FIXME: Embed detection
        'url': 'https://www.punknews.org/article/85809/stay-inside-super-sonic',
        'info_dict': {
            'id': '2475540375',
            'ext': 'mp3',
            'title': 'Stay Inside - Super Sonic',
            'album': 'Lunger',
            'album_artists': ['Stay Inside'],
            'artists': ['Stay Inside'],
            'duration': 166.157,
            'release_date': '20251003',
            'release_timestamp': 1759449600.0,
            'thumbnail': r're:https?://f4\.bcbits\.com/img/.+\.jpg',
            'timestamp': 1749473029.0,
            'track': 'Super Sonic',
            'track_id': '2475540375',
            'track_number': 3,
            'upload_date': '20250609',
            'uploader': 'Stay Inside',
            'uploader_id': 'stayinside',
            'uploader_url': 'https://stayinside.bandcamp.com',
        },
    }]

    def _extract_data_attr(self, webpage, video_id, attr='tralbum', fatal=True):
        return self._parse_json(self._html_search_regex(
            rf'data-{attr}=(["\'])({{.+?}})\1', webpage,
            attr + ' data', group=2), video_id, fatal=fatal)

    def _real_extract(self, url):
        title, uploader = self._match_valid_url(url).group('id', 'uploader')
        webpage = self._download_webpage(url, title)
        tralbum = self._extract_data_attr(webpage, title)
        thumbnail = self._og_search_thumbnail(webpage)

        track_id = None
        track = None
        track_number = None
        duration = None

        formats = []
        track_info = try_get(tralbum, lambda x: x['trackinfo'][0], dict)
        if track_info:
            file_ = track_info.get('file')
            if isinstance(file_, dict):
                for format_id, format_url in file_.items():
                    if not url_or_none(format_url):
                        continue
                    ext, abr_str = format_id.split('-', 1)
                    formats.append({
                        'format_id': format_id,
                        'url': self._proto_relative_url(format_url, 'http:'),
                        'ext': ext,
                        'vcodec': 'none',
                        'acodec': ext,
                        'abr': int_or_none(abr_str),
                    })
            track = track_info.get('title')
            track_id = str_or_none(
                track_info.get('track_id') or track_info.get('id'))
            track_number = int_or_none(track_info.get('track_num'))
            duration = float_or_none(track_info.get('duration'))

        embed = self._extract_data_attr(webpage, title, 'embed', False)
        current = tralbum.get('current') or {}
        artist = embed.get('artist') or current.get('artist') or tralbum.get('artist')
        album_artist = self._html_search_regex(
            r'<h3 class="albumTitle">[\S\s]*?by\s*<span>\s*<a href="[^>]+">\s*([^>]+?)\s*</a>',
            webpage, 'album artist', fatal=False)
        timestamp = unified_timestamp(
            current.get('publish_date') or tralbum.get('album_publish_date'))

        download_link = tralbum.get('freeDownloadPage')
        if download_link:
            track_id = str(tralbum['id'])

            download_webpage = self._download_webpage(
                download_link, track_id, 'Downloading free downloads page')

            blob = self._extract_data_attr(download_webpage, track_id, 'blob')

            info = try_get(
                blob, (lambda x: x['digital_items'][0],
                       lambda x: x['download_items'][0]), dict)
            if info:
                downloads = info.get('downloads')
                if isinstance(downloads, dict):
                    if not track:
                        track = info.get('title')
                    if not artist:
                        artist = info.get('artist')
                    if not thumbnail:
                        thumbnail = info.get('thumb_url')

                    download_formats = {}
                    download_formats_list = blob.get('download_formats')
                    if isinstance(download_formats_list, list):
                        for f in blob['download_formats']:
                            name, ext = f.get('name'), f.get('file_extension')
                            if all(isinstance(x, str) for x in (name, ext)):
                                download_formats[name] = ext.strip('.')

                    for format_id, f in downloads.items():
                        format_url = f.get('url')
                        if not format_url:
                            continue
                        # Stat URL generation algorithm is reverse engineered from
                        # download_*_bundle_*.js
                        stat_url = update_url_query(
                            format_url.replace('/download/', '/statdownload/'), {
                                '.rand': int(time.time() * 1000 * random.random()),
                            })
                        format_id = f.get('encoding_name') or format_id
                        stat = self._download_json(
                            stat_url, track_id, f'Downloading {format_id} JSON',
                            transform_source=lambda s: s[s.index('{'):s.rindex('}') + 1],
                            fatal=False)
                        if not stat:
                            continue
                        retry_url = url_or_none(stat.get('retry_url'))
                        if not retry_url:
                            continue
                        formats.append({
                            'url': self._proto_relative_url(retry_url, 'http:'),
                            'ext': download_formats.get(format_id),
                            'format_id': format_id,
                            'format_note': f.get('description'),
                            'filesize': parse_filesize(f.get('size_mb')),
                            'vcodec': 'none',
                            'acodec': format_id.split('-')[0],
                        })

        title = f'{artist} - {track}' if artist else track

        if not duration:
            duration = float_or_none(self._html_search_meta(
                'duration', webpage, default=None))

        return {
            'id': track_id,
            'title': title,
            'thumbnail': thumbnail,
            'uploader': artist,
            'uploader_id': uploader,
            'uploader_url': f'https://{uploader}.bandcamp.com',
            'timestamp': timestamp,
            'release_timestamp': unified_timestamp(tralbum.get('album_release_date')),
            'duration': duration,
            'track': track,
            'track_number': track_number,
            'track_id': track_id,
            'artist': artist,
            'album': embed.get('album_title'),
            'album_artist': album_artist,
            'formats': formats,
            'tags': traverse_obj(webpage, ({find_elements(cls='tag')}, ..., {clean_html})),
        }


class BandcampAlbumIE(BandcampIE):  # XXX: Do not subclass from concrete IE
    IE_NAME = 'Bandcamp:album'
    _VALID_URL = r'https?://(?:(?P<subdomain>[^.]+)\.)?bandcamp\.com/album/(?P<id>[^/?#&]+)'

    _TESTS = [{
        'url': 'http://blazo.bandcamp.com/album/jazz-format-mixtape-vol-1',
        'playlist': [
            {
                'md5': '39bc1eded3476e927c724321ddf116cf',
                'info_dict': {
                    'id': '1353101989',
                    'ext': 'mp3',
                    'title': 'Blazo - Intro',
                    'thumbnail': r're:https?://f4\.bcbits\.com/img/.+\.jpg',
                    'timestamp': 1311756226,
                    'upload_date': '20110727',
                    'uploader': 'Blazo',
                    'album_artists': ['Blazo'],
                    'uploader_url': 'https://blazo.bandcamp.com',
                    'release_date': '20110727',
                    'release_timestamp': 1311724800.0,
                    'track': 'Intro',
                    'uploader_id': 'blazo',
                    'track_number': 1,
                    'album': 'Jazz Format Mixtape vol.1',
                    'artists': ['Blazo'],
                    'duration': 19.335,
                    'track_id': '1353101989',
                },
            },
            {
                'md5': '1a2c32e2691474643e912cc6cd4bffaa',
                'info_dict': {
                    'id': '38097443',
                    'ext': 'mp3',
                    'title': 'Blazo - Kero One - Keep It Alive (Blazo remix)',
                    'thumbnail': r're:https?://f4\.bcbits\.com/img/.+\.jpg',
                    'timestamp': 1311757238,
                    'upload_date': '20110727',
                    'uploader': 'Blazo',
                    'track': 'Kero One - Keep It Alive (Blazo remix)',
                    'release_date': '20110727',
                    'track_id': '38097443',
                    'track_number': 2,
                    'duration': 181.467,
                    'uploader_url': 'https://blazo.bandcamp.com',
                    'album': 'Jazz Format Mixtape vol.1',
                    'uploader_id': 'blazo',
                    'album_artists': ['Blazo'],
                    'artists': ['Blazo'],
                    'release_timestamp': 1311724800.0,
                },
            },
        ],
        'info_dict': {
            'title': 'Jazz Format Mixtape vol.1',
            'id': 'jazz-format-mixtape-vol-1',
            'uploader_id': 'blazo',
            'description': 'md5:38052a93217f3ffdc033cd5dbbce2989',
        },
        'params': {
            'playlistend': 2,
        },
        'skip': 'Bandcamp imposes download limits.',
    }, {
        'url': 'http://nightbringer.bandcamp.com/album/hierophany-of-the-open-grave',
        'info_dict': {
            'title': 'Hierophany of the Open Grave',
            'uploader_id': 'nightbringer',
            'id': 'hierophany-of-the-open-grave',
        },
        'playlist_mincount': 9,
    }, {
        # with escaped quote in title
        'url': 'https://jstrecords.bandcamp.com/album/entropy-ep',
        'info_dict': {
            'title': '"Entropy" EP',
            'uploader_id': 'jstrecords',
            'id': 'entropy-ep',
            'description': 'md5:0ff22959c943622972596062f2f366a5',
        },
        'playlist_mincount': 3,
    }, {
        # not all tracks have songs
        'url': 'https://insulters.bandcamp.com/album/we-are-the-plague',
        'info_dict': {
            'id': 'we-are-the-plague',
            'title': 'WE ARE THE PLAGUE',
            'uploader_id': 'insulters',
            'description': 'md5:b3cf845ee41b2b1141dc7bde9237255f',
        },
        'playlist_count': 2,
    }]

    @classmethod
    def suitable(cls, url):
        return (False
                if BandcampWeeklyIE.suitable(url) or BandcampIE.suitable(url)
                else super().suitable(url))

    def _real_extract(self, url):
        uploader_id, album_id = self._match_valid_url(url).groups()
        playlist_id = album_id or uploader_id
        webpage = self._download_webpage(url, playlist_id)
        tralbum = self._extract_data_attr(webpage, playlist_id)
        track_info = tralbum.get('trackinfo')
        if not track_info:
            raise ExtractorError('The page doesn\'t contain any tracks')
        # Only tracks with duration info have songs
        entries = [
            self.url_result(
                urljoin(url, t['title_link']), BandcampIE.ie_key(),
                str_or_none(t.get('track_id') or t.get('id')), t.get('title'))
            for t in track_info
            if t.get('duration')]

        current = tralbum.get('current') or {}

        return {
            '_type': 'playlist',
            'uploader_id': uploader_id,
            'id': playlist_id,
            'title': current.get('title'),
            'description': current.get('about'),
            'entries': entries,
        }


class BandcampWeeklyIE(BandcampIE):  # XXX: Do not subclass from concrete IE
    IE_NAME = 'Bandcamp:weekly'
    _VALID_URL = r'https?://(?:www\.)?bandcamp\.com/radio/?\?(?:[^#]+&)?show=(?P<id>\d+)'
    _TESTS = [{
        'url': 'https://bandcamp.com/radio?show=224',
        'md5': '61acc9a002bed93986b91168aa3ab433',
        'info_dict': {
            'id': '224',
            'ext': 'mp3',
            'title': 'Bandcamp Weekly, 2017-04-04',
            'description': 'md5:5d48150916e8e02d030623a48512c874',
            'thumbnail': 'https://f4.bcbits.com/img/9982549_0.jpg',
            'series': 'Bandcamp Weekly',
            'episode_id': '224',
            'release_timestamp': 1491264000,
            'release_date': '20170404',
            'duration': 5829.77,
        },
        'params': {
            'format': 'mp3-128',
        },
    }, {
        'url': 'https://bandcamp.com/radio/?foo=bar&show=224',
        'only_matching': True,
    }]

    def _real_extract(self, url):
        show_id = self._match_id(url)
        show_data = self._download_json(
            'https://bandcamp.com/api/bcradio_api/1/get_show',
            show_id, 'Downloading radio show JSON',
            data=json.dumps({'id': show_id}).encode(),
            headers={'Content-Type': 'application/json'})
        audio_data = show_data['compiledTrack']

        stream_url = audio_data['streamUrl']
        format_id = traverse_obj(stream_url, ({parse_qs}, 'enc', -1))
        encoding, _, bitrate_str = (format_id or '').partition('-')

        series_title = show_data.get('title')
        release_timestamp = unified_timestamp(show_data.get('date'))

        return {
            'id': show_id,
            'episode_id': show_id,
            'title': join_nonempty(series_title, strftime_or_none(release_timestamp, '%Y-%m-%d'), delim=', '),
            'series': series_title,
            'thumbnail': format_field(show_data, 'imageId', 'https://f4.bcbits.com/img/%s_0.jpg', default=None),
            'description': show_data.get('description'),
            'duration': float_or_none(audio_data.get('duration')),
            'release_timestamp': release_timestamp,
            'formats': [{
                'url': stream_url,
                'format_id': format_id,
                'ext': encoding or 'mp3',
                'acodec': encoding or None,
                'vcodec': 'none',
                'abr': int_or_none(bitrate_str),
            }],
        }


class BandcampUserIE(InfoExtractor):
    IE_NAME = 'Bandcamp:user'
    _VALID_URL = r'https?://(?!www\.)(?P<id>[^.]+)\.bandcamp\.com(?:/music)?/?(?:[#?]|$)'

    _TESTS = [{
        # Type 1 Bandcamp user page.
        'url': 'https://adrianvonziegler.bandcamp.com',
        'info_dict': {
            'id': 'adrianvonziegler',
            'title': 'Discography of adrianvonziegler',
        },
        'playlist_mincount': 23,
    }, {
        # Bandcamp user page with only one album
        'url': 'http://dotscale.bandcamp.com',
        'info_dict': {
            'id': 'dotscale',
            'title': 'Discography of dotscale',
        },
        'playlist_count': 1,
    }, {
        # Type 2 Bandcamp user page.
        'url': 'https://nightcallofficial.bandcamp.com',
        'info_dict': {
            'id': 'nightcallofficial',
            'title': 'Discography of nightcallofficial',
        },
        'playlist_count': 4,
    }, {
        'url': 'https://steviasphere.bandcamp.com/music',
        'playlist_mincount': 47,
        'info_dict': {
            'id': 'steviasphere',
            'title': 'Discography of steviasphere',
        },
    }, {
        'url': 'https://coldworldofficial.bandcamp.com/music',
        'playlist_mincount': 7,
        'info_dict': {
            'id': 'coldworldofficial',
            'title': 'Discography of coldworldofficial',
        },
    }, {
        'url': 'https://nuclearwarnowproductions.bandcamp.com/music',
        'playlist_mincount': 399,
        'info_dict': {
            'id': 'nuclearwarnowproductions',
            'title': 'Discography of nuclearwarnowproductions',
        },
    }]

    def _yield_items(self, webpage):
        yield from (
            re.findall(r'<li data-item-id=["\'][^>]+>\s*<a href=["\'](?![^"\'/]*?/merch)([^"\']+)', webpage)
            or re.findall(r'<div[^>]+trackTitle["\'][^"\']+["\']([^"\']+)', webpage))

        yield from traverse_obj(webpage, (
            {find_element(id='music-grid', html=True)}, {extract_attributes},
            'data-client-items', {json.loads}, ..., 'page_url', {str}))

    def _real_extract(self, url):
        uploader = self._match_id(url)
        webpage = self._download_webpage(url, uploader)

        return self.playlist_from_matches(
            self._yield_items(webpage), uploader, f'Discography of {uploader}',
            getter=urljoin(url))



import itertools

from .common import InfoExtractor
from ..utils import (
    clean_html,
    extract_attributes,
    get_element_by_class,
    get_element_html_by_class,
    get_elements_html_by_class,
    parse_qs,
    traverse_obj,
    unified_strdate,
    urljoin,
)


class TheGuardianPodcastIE(InfoExtractor):
    _VALID_URL = r'https?://(?:www\.)?theguardian\.com/\w+/audio/\d{4}/\w{3}/\d{1,2}/(?P<id>[\w-]+)'
    _TESTS = [{
        'url': 'https://www.theguardian.com/news/audio/2023/nov/03/we-are-just-getting-started-the-plastic-eating-bacteria-that-could-change-the-world-podcast',
        'md5': 'd1771744681789b4cd7da2a08e487702',
        'info_dict': {
            'id': 'we-are-just-getting-started-the-plastic-eating-bacteria-that-could-change-the-world-podcast',
            'ext': 'mp3',
            'title': '‘We are just getting started’: the plastic-eating bacteria that could change the world – podcast',
            'description': 'md5:cfd3df2791d394d2ab62cd571d5207ee',
            'creator': 'Stephen Buranyi',
            'thumbnail': 'md5:73c12558fcb3b0e2a59422bfb33b3f79',
            'release_date': '20231103',
        },
    }, {
        'url': 'https://www.theguardian.com/news/audio/2023/oct/30/the-trials-of-robert-habeck-is-the-worlds-most-powerful-green-politician-doomed-to-fail-podcast',
        'md5': 'd1771744681789b4cd7da2a08e487702',
        'info_dict': {
            'id': 'the-trials-of-robert-habeck-is-the-worlds-most-powerful-green-politician-doomed-to-fail-podcast',
            'ext': 'mp3',
            'title': 'The trials of Robert Habeck: is the world’s most powerful green politician doomed to fail? – podcast',
            'description': 'md5:1b5cf6582d1771c6b7077784b5456994',
            'creator': 'Philip Oltermann',
            'thumbnail': 'md5:6e5c5ec43843e956e20be793722e9080',
            'release_date': '20231030',
        },
    }, {
        'url': 'https://www.theguardian.com/football/audio/2023/nov/06/arsenal-feel-hard-done-by-and-luton-hold-liverpool-football-weekly',
        'md5': 'a2fcff6f8e060a95b1483295273dc35e',
        'info_dict': {
            'id': 'arsenal-feel-hard-done-by-and-luton-hold-liverpool-football-weekly',
            'ext': 'mp3',
            'title': 'Arsenal feel hard done by and Luton hold Liverpool – Football Weekly',
            'description': 'md5:286a9fbddaeb7c83cc65d1c4a5330b2a',
            'creator': 'Max Rushden',
            'thumbnail': 'md5:93eb7d6440f1bb94eb3a6cad63f48afd',
            'release_date': '20231106',
        },
    }, {
        'url': 'https://www.theguardian.com/politics/audio/2023/nov/02/the-covid-inquiry-politics-weekly-uk-podcast',
        'md5': '06a0f7e9701a80c8064a5d35690481ec',
        'info_dict': {
            'id': 'the-covid-inquiry-politics-weekly-uk-podcast',
            'ext': 'mp3',
            'title': 'The Covid inquiry | Politics Weekly UK - podcast',
            'description': 'md5:207c98859c14903582b17d25b014046e',
            'creator': 'Gaby Hinsliff',
            'thumbnail': 'md5:28932a7b5a25b057be330d2ed70ea7f3',
            'release_date': '20231102',
        },
    }]

    def _real_extract(self, url):
        video_id = self._match_id(url)
        webpage = self._download_webpage(url, video_id)
        return {
            'id': video_id,
            'title': self._og_search_title(webpage) or get_element_by_class('content__headline', webpage),
            'description': self._og_search_description(webpage),
            'creator': self._html_search_meta('author', webpage),
            'thumbnail': self._og_search_thumbnail(webpage),
            'release_date': unified_strdate(self._html_search_meta('article:published_time', webpage)),
            'url': extract_attributes(get_element_html_by_class(
                'podcast__player', webpage) or '').get('data-source'),
        }


class TheGuardianPodcastPlaylistIE(InfoExtractor):
    _VALID_URL = r'https?://(?:www\.)?theguardian\.com/\w+/series/(?P<id>[\w-]+)(?:\?page=\d+)?'
    _TESTS = [{
        'url': 'https://www.theguardian.com/football/series/theguardianswomensfootballweekly',
        'info_dict': {
            'id': 'theguardianswomensfootballweekly',
            'title': "The Guardian's Women's Football Weekly",
            'description': 'md5:e2cc021311e582d29935a73614a43f51',
        },
        'playlist_mincount': 69,
    }, {
        'url': 'https://www.theguardian.com/news/series/todayinfocus?page=2',
        'info_dict': {
            'id': 'todayinfocus',
            'title': 'Today in Focus',
            'description': 'md5:0f097764fc0d359e0b6eb537be0387e2',
        },
        'playlist_mincount': 1261,
    }, {
        'url': 'https://www.theguardian.com/news/series/the-audio-long-read',
        'info_dict': {
            'id': 'the-audio-long-read',
            'title': 'The Audio Long Read',
            'description': 'md5:5462994a27527309562b25b6defc4ef3',
        },
        'playlist_mincount': 996,
    }]

    def _entries(self, url, playlist_id):
        for page in itertools.count(1):
            webpage, urlh = self._download_webpage_handle(
                url, playlist_id, f'Downloading page {page}', query={'page': page})
            if 'page' not in parse_qs(urlh.url):
                break

            episodes = get_elements_html_by_class('fc-item--type-media', webpage)
            yield from traverse_obj(episodes, (..., {extract_attributes}, 'data-id'))

    def _real_extract(self, url):
        podcast_id = self._match_id(url)

        webpage = self._download_webpage(url, podcast_id)

        title = clean_html(get_element_by_class(
            'index-page-header__title', webpage) or get_element_by_class('flagship-audio__title', webpage))
        description = self._og_search_description(webpage) or self._html_search_meta(
            'description', webpage)

        return self.playlist_from_matches(
            self._entries(url, podcast_id), podcast_id, title, description=description,
            ie=TheGuardianPodcastIE, getter=urljoin('https://www.theguardian.com'))



import re

from .common import InfoExtractor
from ..utils import (
    ExtractorError,
    get_element_by_attribute,
)


class TMZIE(InfoExtractor):
    _VALID_URL = r'https?://(?:www\.)?tmz\.com/.*'
    _TESTS = [
        {
            'url': 'http://www.tmz.com/videos/0-cegprt2p/',
            'info_dict': {
                'id': 'http://www.tmz.com/videos/0-cegprt2p/',
                'ext': 'mp4',
                'title': 'No Charges Against Hillary Clinton? Harvey Says It Ain\'t Over Yet',
                'description': 'Harvey talks about Director Comey’s decision not to prosecute Hillary Clinton.',
                'timestamp': 1467831837,
                'uploader': 'TMZ Staff',
                'upload_date': '20160706',
                'thumbnail': 'https://imagez.tmz.com/image/5e/4by3/2016/07/06/5eea7dc01baa5c2e83eb06930c170e46_xl.jpg',
                'duration': 772.0,
            },
        },
        {
            'url': 'https://www.tmz.com/videos/071119-chris-morgan-women-4590005-0-zcsejvcr/',
            'info_dict': {
                'id': 'https://www.tmz.com/videos/071119-chris-morgan-women-4590005-0-zcsejvcr/',
                'ext': 'mp4',
                'title': 'Angry Bagel Shop Guy Says He Doesn\'t Trust Women',
                'description': 'The enraged man who went viral for ranting about women on dating sites before getting ragdolled in a bagel shop is defending his misogyny ... he says it\'s women\'s fault in the first place.',
                'timestamp': 1562889485,
                'uploader': 'TMZ Staff',
                'upload_date': '20190711',
                'thumbnail': 'https://imagez.tmz.com/image/a8/4by3/2019/07/12/a85480d27b2f50a7bfea2322151d67a5_xl.jpg',
                'duration': 123.0,
            },
        },
        {
            'url': 'http://www.tmz.com/2015/04/19/bobby-brown-bobbi-kristina-awake-video-concert',
            'md5': '5429c85db8bde39a473a56ca8c4c5602',
            'info_dict': {
                'id': 'http://www.tmz.com/2015/04/19/bobby-brown-bobbi-kristina-awake-video-concert',
                'ext': 'mp4',
                'title': 'Bobby Brown Tells Crowd ... Bobbi Kristina is Awake',
                'description': 'Bobby Brown stunned his audience during a concert Saturday night, when he told the crowd, "Bobbi is awake.  She\'s watching me."',
                'timestamp': 1429467813,
                'uploader': 'TMZ Staff',
                'upload_date': '20150419',
                'duration': 29.0,
                'thumbnail': 'https://imagez.tmz.com/image/15/4by3/2015/04/20/1539c7ae136359fc979236fa6a9449dd_xl.jpg',
            },
        },
        {
            'url': 'http://www.tmz.com/2015/09/19/patti-labelle-concert-fan-stripping-kicked-out-nicki-minaj/',
            'info_dict': {
                'id': 'http://www.tmz.com/2015/09/19/patti-labelle-concert-fan-stripping-kicked-out-nicki-minaj/',
                'ext': 'mp4',
                'title': 'Patti LaBelle -- Goes Nuclear On Stripping Fan',
                'description': 'Patti LaBelle made it known loud and clear last night ... NO '
                'ONE gets on her stage and strips down.',
                'timestamp': 1442683746,
                'uploader': 'TMZ Staff',
                'upload_date': '20150919',
                'duration': 104.0,
                'thumbnail': 'https://imagez.tmz.com/image/5e/4by3/2015/09/20/5e57d7575062528082994e18ac3f0f48_xl.jpg',
            },
        },
        {
            'url': 'http://www.tmz.com/2016/01/28/adam-silver-sting-drake-blake-griffin/',
            'info_dict': {
                'id': 'http://www.tmz.com/2016/01/28/adam-silver-sting-drake-blake-griffin/',
                'ext': 'mp4',
                'title': 'NBA\'s Adam Silver -- Blake Griffin\'s a Great Guy ... He\'ll Learn from This',
                'description': 'Two pretty parts of this video with NBA Commish Adam Silver.',
                'timestamp': 1454010989,
                'uploader': 'TMZ Staff',
                'upload_date': '20160128',
                'duration': 59.0,
                'thumbnail': 'https://imagez.tmz.com/image/38/4by3/2016/01/29/3856e83e0beb57059ec412122b842fb1_xl.jpg',
            },
        },
        {
            'url': 'http://www.tmz.com/2016/10/27/donald-trump-star-vandal-arrested-james-otis/',
            'info_dict': {
                'id': 'http://www.tmz.com/2016/10/27/donald-trump-star-vandal-arrested-james-otis/',
                'ext': 'mp4',
                'title': 'Trump Star Vandal -- I\'m Not Afraid of Donald or the Cops!',
                'description': 'James Otis is the the guy who took a pickaxe to Donald Trump\'s star on the Walk of Fame, and he tells TMZ .. he\'s ready and willing to go to jail for the crime.',
                'timestamp': 1477500095,
                'uploader': 'TMZ Staff',
                'upload_date': '20161026',
                'thumbnail': 'https://imagez.tmz.com/image/0d/4by3/2016/10/27/0d904814d4a75dcf9cc3b8cfd1edc1a3_xl.jpg',
                'duration': 128.0,
            },
        },
        {
            'url': 'https://www.tmz.com/videos/2020-10-31-103120-beverly-hills-protest-4878209/',
            'info_dict': {
                'id': 'https://www.tmz.com/videos/2020-10-31-103120-beverly-hills-protest-4878209/',
                'ext': 'mp4',
                'title': 'Cops Use Billy Clubs Against Pro-Trump and Anti-Fascist '
                'Demonstrators',
                'description': 'Beverly Hills may be an omen of what\'s coming next week, '
                'because things got crazy on the streets and cops started '
                'swinging their billy clubs at both Anti-Fascist and Pro-Trump '
                'demonstrators.',
                'timestamp': 1604182772,
                'uploader': 'TMZ Staff',
                'upload_date': '20201031',
                'duration': 96.0,
                'thumbnail': 'https://imagez.tmz.com/image/f3/4by3/2020/10/31/f37bd5a8aef84497866f425130c58be3_xl.jpg',
            },
        },
        {
            'url': 'https://www.tmz.com/2020/11/05/gervonta-davis-car-crash-hit-and-run-police/',
            'info_dict': {
                'id': 'Dddb6IGe-ws',
                'ext': 'mp4',
                'title': 'SICK LAMBO GERVONTA DAVIS IN HIS NEW RIDE RIGHT AFTER KO AFTER LEO  EsNews Boxing',
                'uploader': 'ESNEWS',
                'description': 'md5:49675bc58883ccf80474b8aa701e1064',
                'upload_date': '20201102',
                'uploader_id': '@ESNEWS',
                'uploader_url': 'https://www.youtube.com/@ESNEWS',
                'like_count': int,
                'channel_id': 'UCI-Oq7oFGakzSzHFlTtsUsQ',
                'channel': 'ESNEWS',
                'view_count': int,
                'duration': 225,
                'live_status': 'not_live',
                'thumbnail': 'https://i.ytimg.com/vi_webp/Dddb6IGe-ws/maxresdefault.webp',
                'channel_url': 'https://www.youtube.com/channel/UCI-Oq7oFGakzSzHFlTtsUsQ',
                'channel_follower_count': int,
                'playable_in_embed': True,
                'categories': ['Sports'],
                'age_limit': 0,
                'tags': 'count:10',
                'availability': 'public',
                'comment_count': int,
            },
        },
        {
            'url': 'https://www.tmz.com/2020/11/19/conor-mcgregor-dustin-poirier-contract-fight-ufc-257-fight-island/',
            'info_dict': {
                'id': '1329448013937471491',
                'ext': 'mp4',
                'title': 'The Mac Life - BREAKING: Conor McGregor (@thenotoriousmma) has signed his bout agreement for his rematch with Dustin Poirier for January 23.',
                'uploader': 'The Mac Life',
                'description': 'md5:56e6009bbc3d12498e10d08a8e1f1c69',
                'upload_date': '20201119',
                'display_id': '1329450007125225473',
                'uploader_id': 'TheMacLife',
                'timestamp': 1605800556,
                'thumbnail': 'https://pbs.twimg.com/media/EnMmfT8XYAExgxJ.jpg?name=small',
                'like_count': int,
                'duration': 11.812,
                'uploader_url': 'https://twitter.com/TheMacLife',
                'age_limit': 0,
                'repost_count': int,
                'tags': [],
                'comment_count': int,
            },
        },
    ]

    def _real_extract(self, url):
        webpage = self._download_webpage(url, url)
        jsonld = self._search_json_ld(webpage, url)
        if not jsonld or 'url' not in jsonld:
            # try to extract from YouTube Player API
            # see https://developers.google.com/youtube/iframe_api_reference#Video_Queueing_Functions
            match_obj = re.search(r'\.cueVideoById\(\s*(?P<quote>[\'"])(?P<id>.*?)(?P=quote)', webpage)
            if match_obj:
                return self.url_result(match_obj.group('id'))
            # try to extract from twitter
            blockquote_el = get_element_by_attribute('class', 'twitter-tweet', webpage)
            if blockquote_el:
                matches = re.findall(
                    r'<a[^>]+href=\s*(?P<quote>[\'"])(?P<link>.*?)(?P=quote)',
                    blockquote_el)
                if matches:
                    for _, match in matches:
                        if '/status/' in match:
                            return self.url_result(match)
            raise ExtractorError('No video found!')
        if id not in jsonld:
            jsonld['id'] = url
        return jsonld



import json

from .common import InfoExtractor
from ..utils import (
    determine_ext,
    float_or_none,
    int_or_none,
    str_or_none,
    strip_or_none,
    traverse_obj,
    unified_timestamp,
    url_or_none,
)


class PinterestBaseIE(InfoExtractor):
    _VALID_URL_BASE = r'''(?x)
        https?://(?:[^/]+\.)?pinterest\.(?:
            com|fr|de|ch|jp|cl|ca|it|co\.uk|nz|ru|com\.au|at|pt|co\.kr|es|com\.mx|
            dk|ph|th|com\.uy|co|nl|info|kr|ie|vn|com\.vn|ec|mx|in|pe|co\.at|hu|
            co\.in|co\.nz|id|com\.ec|com\.py|tw|be|uk|com\.bo|com\.pe)'''

    def _call_api(self, resource, video_id, options):
        return self._download_json(
            f'https://www.pinterest.com/resource/{resource}Resource/get/',
            video_id, f'Download {resource} JSON metadata',
            query={'data': json.dumps({'options': options})},
            headers={'X-Pinterest-PWS-Handler': 'www/[username].js'})['resource_response']

    def _extract_video(self, data, extract_formats=True):
        video_id = data['id']
        thumbnails = []
        images = data.get('images')
        if isinstance(images, dict):
            for thumbnail in images.values():
                if not isinstance(thumbnail, dict):
                    continue
                thumbnail_url = url_or_none(thumbnail.get('url'))
                if not thumbnail_url:
                    continue
                thumbnails.append({
                    'url': thumbnail_url,
                    'width': int_or_none(thumbnail.get('width')),
                    'height': int_or_none(thumbnail.get('height')),
                })

        info = {
            'title': strip_or_none(traverse_obj(data, 'title', 'grid_title', default='')),
            'description': traverse_obj(data, 'seo_description', 'description'),
            'timestamp': unified_timestamp(data.get('created_at')),
            'thumbnails': thumbnails,
            'uploader': traverse_obj(data, ('closeup_attribution', 'full_name')),
            'uploader_id': str_or_none(traverse_obj(data, ('closeup_attribution', 'id'))),
            'repost_count': int_or_none(data.get('repin_count')),
            'comment_count': int_or_none(data.get('comment_count')),
            'categories': traverse_obj(data, ('pin_join', 'visual_annotation'), expected_type=list),
            'tags': traverse_obj(data, 'hashtags', expected_type=list),
        }

        urls = []
        formats = []
        duration = None
        domain = data.get('domain', '')
        if domain.lower() != 'uploaded by user' and traverse_obj(data, ('embed', 'src')):
            if not info['title']:
                info['title'] = None
            return {
                '_type': 'url_transparent',
                'url': data['embed']['src'],
                **info,
            }

        elif extract_formats:
            video_list = traverse_obj(
                data, ('videos', 'video_list'),
                ('story_pin_data', 'pages', ..., 'blocks', ..., 'video', 'video_list'),
                expected_type=dict, get_all=False, default={})
            for format_id, format_dict in video_list.items():
                if not isinstance(format_dict, dict):
                    continue
                format_url = url_or_none(format_dict.get('url'))
                if not format_url or format_url in urls:
                    continue
                urls.append(format_url)
                duration = float_or_none(format_dict.get('duration'), scale=1000)
                ext = determine_ext(format_url)
                if 'hls' in format_id.lower() or ext == 'm3u8':
                    formats.extend(self._extract_m3u8_formats(
                        format_url, video_id, 'mp4', entry_protocol='m3u8_native',
                        m3u8_id=format_id, fatal=False))
                else:
                    formats.append({
                        'url': format_url,
                        'format_id': format_id,
                        'width': int_or_none(format_dict.get('width')),
                        'height': int_or_none(format_dict.get('height')),
                        'duration': duration,
                    })

        return {
            'id': video_id,
            'formats': formats,
            'duration': duration,
            'webpage_url': f'https://www.pinterest.com/pin/{video_id}/',
            'extractor_key': PinterestIE.ie_key(),
            'extractor': PinterestIE.IE_NAME,
            **info,
        }


class PinterestIE(PinterestBaseIE):
    _VALID_URL = rf'{PinterestBaseIE._VALID_URL_BASE}/pin/(?:[\w-]+--)?(?P<id>\d+)'
    _TESTS = [{
        # formats found in data['videos']
        'url': 'https://www.pinterest.com/pin/664281013778109217/',
        'md5': '6550c2af85d6d9f3fe3b88954d1577fc',
        'info_dict': {
            'id': '664281013778109217',
            'ext': 'mp4',
            'title': 'Origami',
            'description': 'md5:e29801cab7d741ea8c741bc50c8d00ab',
            'duration': 57.7,
            'timestamp': 1593073622,
            'upload_date': '20200625',
            'repost_count': int,
            'comment_count': int,
            'categories': list,
            'tags': list,
            'thumbnail': r're:^https?://.*\.(?:jpg|png)$',
        },
    }, {
        # formats found in data['story_pin_data']
        'url': 'https://www.pinterest.com/pin/1084663891475263837/',
        'md5': '069ac19919ab9e1e13fa60de46290b03',
        'info_dict': {
            'id': '1084663891475263837',
            'ext': 'mp4',
            'title': 'Gadget, Cool products, Amazon product, technology, Kitchen gadgets',
            'description': 'md5:d0a4b6ae996ff0c6eed83bc869598d13',
            'uploader': 'CoolCrazyGadgets',
            'uploader_id': '1084664028912989237',
            'upload_date': '20211003',
            'timestamp': 1633246654.0,
            'duration': 14.9,
            'comment_count': int,
            'repost_count': int,
            'categories': 'count:9',
            'tags': list,
            'thumbnail': r're:^https?://.*\.(?:jpg|png)$',
        },
    }, {
        # vimeo.com embed
        'url': 'https://www.pinterest.ca/pin/441282463481903715/',
        'info_dict': {
            'id': '111691128',
            'ext': 'mp4',
            'title': 'Tonite Let\'s All Make Love In London (1967)',
            'description': 'md5:8190f37b3926807809ec57ec21aa77b2',
            'uploader': 'Vimeo',
            'uploader_id': '473792960706651251',
            'upload_date': '20180120',
            'timestamp': 1516409040,
            'duration': 3404,
            'comment_count': int,
            'repost_count': int,
            'categories': 'count:9',
            'tags': [],
            'thumbnail': r're:^https?://.*\.(?:jpg|png)$',
            'uploader_url': 'https://vimeo.com/willardandrade',
        },
        'params': {
            'skip_download': 'm3u8',
        },
    }, {
        'url': 'https://co.pinterest.com/pin/824721750502199491/',
        'only_matching': True,
    },
        {
        'url': 'https://pinterest.com/pin/dive-into-serenity-blue-lagoon-pedi-nails-for-a-tranquil-and-refreshing-spa-experience-video-in-2024--2885187256207927',
        'info_dict': {
            'id': '2885187256207927',
            'ext': 'mp4',
            'title': 'Dive into Serenity: Blue Lagoon Pedi Nails for a Tranquil and Refreshing Spa Experience! 💙💅',
            'description': 'md5:5da41c767d2317e42e49b663b0b2150f',
            'uploader': 'Glamour Artistry |Everyday Outfits, Luxury Fashion & Nail Designs',
            'uploader_id': '1142999717836434688',
            'upload_date': '20240702',
            'timestamp': 1719939156,
            'duration': 7.967,
            'comment_count': int,
            'repost_count': int,
            'categories': 'count:9',
            'tags': ['#BlueLagoonPediNails', '#SpaExperience'],
            'thumbnail': r're:^https?://.*\.(?:jpg|png)$',
        },
    }]

    def _real_extract(self, url):
        video_id = self._match_id(url)
        data = self._call_api(
            'Pin', video_id, {
                'field_set_key': 'unauth_react_main_pin',
                'id': video_id,
            })['data']
        return self._extract_video(data)


class PinterestCollectionIE(PinterestBaseIE):
    _VALID_URL = rf'{PinterestBaseIE._VALID_URL_BASE}/(?P<username>[^/]+)/(?P<id>[^/?#&]+)'
    _TESTS = [{
        'url': 'https://www.pinterest.ca/mashal0407/cool-diys/',
        'info_dict': {
            'id': '585890301462791043',
            'title': 'cool diys',
        },
        'playlist_count': 8,
    }, {
        'url': 'https://www.pinterest.ca/fudohub/videos/',
        'info_dict': {
            'id': '682858430939307450',
            'title': 'VIDEOS',
        },
        'playlist_mincount': 365,
        'skip': 'Test with extract_formats=False',
    }]

    @classmethod
    def suitable(cls, url):
        return False if PinterestIE.suitable(url) else super().suitable(url)

    def _real_extract(self, url):
        username, slug = self._match_valid_url(url).groups()
        board = self._call_api(
            'Board', slug, {
                'slug': slug,
                'username': username,
            })['data']
        board_id = board['id']
        options = {
            'board_id': board_id,
            'page_size': 250,
        }
        bookmark = None
        entries = []
        while True:
            if bookmark:
                options['bookmarks'] = [bookmark]
            board_feed = self._call_api('BoardFeed', board_id, options)
            for item in (board_feed.get('data') or []):
                if not isinstance(item, dict) or item.get('type') != 'pin':
                    continue
                video_id = item.get('id')
                if video_id:
                    # Some pins may not be available anonymously via pin URL
                    # video = self._extract_video(item, extract_formats=False)
                    # video.update({
                    #     '_type': 'url_transparent',
                    #     'url': 'https://www.pinterest.com/pin/%s/' % video_id,
                    # })
                    # entries.append(video)
                    entries.append(self._extract_video(item))
            bookmark = board_feed.get('bookmark')
            if not bookmark:
                break
        return self.playlist_result(
            entries, playlist_id=board_id, playlist_title=board.get('name'))



import re

from .common import InfoExtractor
from ..utils import (
    clean_html,
    extract_attributes,
    js_to_json,
    mimetype2ext,
    unified_strdate,
    url_or_none,
    urljoin,
)
from ..utils.traversal import find_element, traverse_obj


class TVNoeIE(InfoExtractor):
    IE_NAME = 'tvnoe'
    IE_DESC = 'Televize Noe'

    _VALID_URL = r'https?://(?:www\.)?tvnoe\.cz/porad/(?P<id>[\w-]+)'
    _TESTS = [{
        'url': 'https://www.tvnoe.cz/porad/43216-outdoor-films-s-mudr-tomasem-kempnym-pomahat-potrebnym-nejen-u-nas',
        'info_dict': {
            'id': '43216-outdoor-films-s-mudr-tomasem-kempnym-pomahat-potrebnym-nejen-u-nas',
            'ext': 'mp4',
            'title': 'Pomáhat potřebným nejen u nás',
            'description': 'md5:78b538ee32f7e881ec23b9c278a0ff3a',
            'release_date': '20250531',
            'series': 'Outdoor Films s MUDr. Tomášem Kempným',
            'thumbnail': r're:https?://www\.tvnoe\.cz/.+\.jpg',
        },
    }, {
        'url': 'https://www.tvnoe.cz/porad/43205-zamysleni-tomase-halika-7-nedele-velikonocni',
        'info_dict': {
            'id': '43205-zamysleni-tomase-halika-7-nedele-velikonocni',
            'ext': 'mp4',
            'title': '7. neděle velikonoční',
            'description': 'md5:6bb9908efc59abe60e1c8c7c0e9bb6cd',
            'release_date': '20250531',
            'series': 'Zamyšlení Tomáše Halíka',
            'thumbnail': r're:https?://www\.tvnoe\.cz/.+\.jpg',
        },
    }]

    def _real_extract(self, url):
        video_id = self._match_id(url)
        webpage = self._download_webpage(url, video_id)
        player = self._search_json(
            r'var\s+INIT_PLAYER\s*=', webpage, 'init player',
            video_id, transform_source=js_to_json)

        formats = []
        for source in traverse_obj(player, ('tracks', ..., lambda _, v: url_or_none(v['src']))):
            src_url = source['src']
            ext = mimetype2ext(source.get('type'))
            if ext == 'm3u8':
                fmts = self._extract_m3u8_formats(
                    src_url, video_id, 'mp4', m3u8_id='hls', fatal=False)
            elif ext == 'mpd':
                fmts = self._extract_mpd_formats(
                    src_url, video_id, mpd_id='dash', fatal=False)
            else:
                self.report_warning(f'Unsupported stream type: {ext}')
                continue
            formats.extend(fmts)

        return {
            'id': video_id,
            'description': clean_html(self._search_regex(
                r'<p\s+class="">(.+?)</p>', webpage, 'description', default=None)),
            'formats': formats,
            **traverse_obj(webpage, {
                'title': ({find_element(tag='h2')}, {clean_html}),
                'release_date': (
                    {clean_html}, {re.compile(r'Premiéra:\s*(\d{1,2}\.\d{1,2}\.\d{4})').findall},
                    ..., {str}, {unified_strdate}, any),
                'series': ({find_element(tag='h1')}, {clean_html}),
                'thumbnail': (
                    {find_element(id='player-live', html=True)}, {extract_attributes},
                    'poster', {urljoin('https://www.tvnoe.cz/')}),
            }),
        }



from __future__ import annotations

import bisect
import threading
import time


class ProgressCalculator:
    # Time to calculate the speed over (seconds)
    SAMPLING_WINDOW = 3
    # Minimum timeframe before to sample next downloaded bytes (seconds)
    SAMPLING_RATE = 0.05
    # Time before showing eta (seconds)
    GRACE_PERIOD = 1

    def __init__(self, initial: int):
        self._initial = initial or 0
        self.downloaded = self._initial

        self.elapsed: float = 0
        self.speed = SmoothValue(0, smoothing=0.7)
        self.eta = SmoothValue(None, smoothing=0.9)

        self._total = 0
        self._start_time = time.monotonic()
        self._last_update = self._start_time

        self._lock = threading.Lock()
        self._thread_sizes: dict[int, int] = {}

        self._times = [self._start_time]
        self._downloaded = [self.downloaded]

    @property
    def total(self):
        return self._total

    @total.setter
    def total(self, value: int | None):
        with self._lock:
            if value is not None and value < self.downloaded:
                value = self.downloaded

            self._total = value

    def thread_reset(self):
        current_thread = threading.get_ident()
        with self._lock:
            self._thread_sizes[current_thread] = 0

    def update(self, size: int | None):
        if not size:
            return

        current_thread = threading.get_ident()

        with self._lock:
            last_size = self._thread_sizes.get(current_thread, 0)
            self._thread_sizes[current_thread] = size
            self._update(size - last_size)

    def _update(self, size: int):
        current_time = time.monotonic()

        self.downloaded += size
        self.elapsed = current_time - self._start_time
        if self.total is not None and self.downloaded > self.total:
            self._total = self.downloaded

        if self._last_update + self.SAMPLING_RATE > current_time:
            return
        self._last_update = current_time

        self._times.append(current_time)
        self._downloaded.append(self.downloaded)

        offset = bisect.bisect_left(self._times, current_time - self.SAMPLING_WINDOW)
        del self._times[:offset]
        del self._downloaded[:offset]
        if len(self._times) < 2:
            self.speed.reset()
            self.eta.reset()
            return

        download_time = current_time - self._times[0]
        if not download_time:
            return

        self.speed.set((self.downloaded - self._downloaded[0]) / download_time)
        if self.total and self.speed.value and self.elapsed > self.GRACE_PERIOD:
            self.eta.set((self.total - self.downloaded) / self.speed.value)
        else:
            self.eta.reset()


class SmoothValue:
    def __init__(self, initial: float | None, smoothing: float):
        self.value = self.smooth = self._initial = initial
        self._smoothing = smoothing

    def set(self, value: float):
        self.value = value
        if self.smooth is None:
            self.smooth = self.value
        else:
            self.smooth = (1 - self._smoothing) * value + self._smoothing * self.smooth

    def reset(self):
        self.value = self.smooth = self._initial



import base64
import json
import re

from .common import InfoExtractor
from ..utils import (
    determine_ext,
    int_or_none,
    mimetype2ext,
    qualities,
    traverse_obj,
    try_get,
    url_or_none,
)


class ImdbIE(InfoExtractor):
    IE_NAME = 'imdb'
    IE_DESC = 'Internet Movie Database trailers'
    _VALID_URL = r'https?://(?:www|m)\.imdb\.com/(?:video|title|list).*?[/-]vi(?P<id>\d+)'

    _TESTS = [{
        'url': 'http://www.imdb.com/video/imdb/vi2524815897',
        'info_dict': {
            'id': '2524815897',
            'ext': 'mp4',
            'title': 'No. 2',
            'description': 'md5:87bd0bdc61e351f21f20d2d7441cb4e7',
            'duration': 152,
            'thumbnail': r're:^https?://.+\.jpg',
        },
    }, {
        'url': 'https://www.imdb.com/video/vi3516832537',
        'info_dict': {
            'id': '3516832537',
            'ext': 'mp4',
            'title': 'Paul: U.S. Trailer #1',
            'description': 'md5:17fcc4fe11ec29b4399be9d4c5ef126c',
            'duration': 153,
            'thumbnail': r're:^https?://.+\.jpg',
        },
    }, {
        'url': 'http://www.imdb.com/video/_/vi2524815897',
        'only_matching': True,
    }, {
        'url': 'http://www.imdb.com/title/tt1667889/?ref_=ext_shr_eml_vi#lb-vi2524815897',
        'only_matching': True,
    }, {
        'url': 'http://www.imdb.com/title/tt1667889/#lb-vi2524815897',
        'only_matching': True,
    }, {
        'url': 'http://www.imdb.com/videoplayer/vi1562949145',
        'only_matching': True,
    }, {
        'url': 'http://www.imdb.com/title/tt4218696/videoplayer/vi2608641561',
        'only_matching': True,
    }, {
        'url': 'https://www.imdb.com/list/ls009921623/videoplayer/vi260482329',
        'only_matching': True,
    }]

    def _real_extract(self, url):
        video_id = self._match_id(url)
        webpage = self._download_webpage(f'https://www.imdb.com/video/vi{video_id}', video_id)
        info = self._search_nextjs_data(webpage, video_id)
        video_info = traverse_obj(info, ('props', 'pageProps', 'videoPlaybackData', 'video'), default={})
        title = (traverse_obj(video_info, ('name', 'value'), ('primaryTitle', 'titleText', 'text'))
                 or self._html_search_meta(('og:title', 'twitter:title'), webpage, default=None)
                 or self._html_extract_title(webpage))
        data = video_info.get('playbackURLs') or try_get(self._download_json(
            'https://www.imdb.com/ve/data/VIDEO_PLAYBACK_DATA', video_id,
            query={
                'key': base64.b64encode(json.dumps({
                    'type': 'VIDEO_PLAYER',
                    'subType': 'FORCE_LEGACY',
                    'id': f'vi{video_id}',
                }).encode()).decode(),
            }), lambda x: x[0]['videoLegacyEncodings'])
        quality = qualities(('SD', '480p', '720p', '1080p'))
        formats, subtitles = [], {}
        for encoding in data:
            if not encoding or not isinstance(encoding, dict):
                continue
            video_url = url_or_none(encoding.get('url'))
            if not video_url:
                continue
            ext = mimetype2ext(encoding.get(
                'mimeType')) or determine_ext(video_url)
            if ext == 'm3u8':
                fmts, subs = self._extract_m3u8_formats_and_subtitles(
                    video_url, video_id, 'mp4', entry_protocol='m3u8_native',
                    preference=1, m3u8_id='hls', fatal=False)
                subtitles = self._merge_subtitles(subtitles, subs)
                formats.extend(fmts)
                continue
            format_id = traverse_obj(encoding, ('displayName', 'value'), 'definition')
            formats.append({
                'format_id': format_id,
                'url': video_url,
                'ext': ext,
                'quality': quality(format_id),
            })

        return {
            'id': video_id,
            'title': title,
            'alt_title': info.get('videoSubTitle'),
            'formats': formats,
            'description': try_get(video_info, lambda x: x['description']['value']),
            'thumbnail': url_or_none(try_get(video_info, lambda x: x['thumbnail']['url'])),
            'duration': int_or_none(try_get(video_info, lambda x: x['runtime']['value'])),
            'subtitles': subtitles,
        }


class ImdbListIE(InfoExtractor):
    IE_NAME = 'imdb:list'
    IE_DESC = 'Internet Movie Database lists'
    _VALID_URL = r'https?://(?:www\.)?imdb\.com/list/ls(?P<id>\d{9})(?!/videoplayer/vi\d+)'
    _TEST = {
        'url': 'https://www.imdb.com/list/ls009921623/',
        'info_dict': {
            'id': '009921623',
            'title': 'The Bourne Legacy',
            'description': 'A list of trailers, clips, and more from The Bourne Legacy, starring Jeremy Renner and Rachel Weisz.',
        },
        'playlist_count': 8,
    }

    def _real_extract(self, url):
        list_id = self._match_id(url)
        webpage = self._download_webpage(url, list_id)
        entries = [
            self.url_result('http://www.imdb.com' + m, 'Imdb')
            for m in re.findall(rf'href="(/list/ls{list_id}/videoplayer/vi[^"]+)"', webpage)]

        list_title = self._html_search_regex(
            r'<h1[^>]+class="[^"]*header[^"]*"[^>]*>(.*?)</h1>',
            webpage, 'list title')
        list_description = self._html_search_regex(
            r'<div[^>]+class="[^"]*list-description[^"]*"[^>]*><p>(.*?)</p>',
            webpage, 'list description')

        return self.playlist_result(entries, list_id, list_title, list_description)



import json
import textwrap
import urllib.parse
import uuid

from .common import InfoExtractor
from ..utils import (
    ExtractorError,
    determine_ext,
    filter_dict,
    int_or_none,
    parse_iso8601,
    update_url,
    url_or_none,
)
from ..utils.traversal import traverse_obj


class LoomIE(InfoExtractor):
    IE_NAME = 'loom'
    _VALID_URL = r'https?://(?:www\.)?loom\.com/(?:share|embed)/(?P<id>[\da-f]{32})'
    _EMBED_REGEX = [rf'<iframe[^>]+\bsrc=["\'](?P<url>{_VALID_URL})']
    _TESTS = [{
        # m3u8 raw-url, mp4 transcoded-url, cdn url == raw-url, json subs only
        'url': 'https://www.loom.com/share/43d05f362f734614a2e81b4694a3a523',
        'md5': 'bfc2d7e9c2e0eb4813212230794b6f42',
        'info_dict': {
            'id': '43d05f362f734614a2e81b4694a3a523',
            'ext': 'mp4',
            'title': 'A Ruler for Windows - 28 March 2022',
            'uploader': 'wILLIAM PIP',
            'upload_date': '20220328',
            'timestamp': 1648454238,
            'duration': 27,
        },
    }, {
        # webm raw-url, mp4 transcoded-url, cdn url == transcoded-url, no subs
        'url': 'https://www.loom.com/share/c43a642f815f4378b6f80a889bb73d8d',
        'md5': '70f529317be8cf880fcc2c649a531900',
        'info_dict': {
            'id': 'c43a642f815f4378b6f80a889bb73d8d',
            'ext': 'webm',
            'title': 'Lilah Nielsen Intro Video',
            'uploader': 'Lilah Nielsen',
            'upload_date': '20200826',
            'timestamp': 1598480716,
            'duration': 20,
        },
    }, {
        # m3u8 raw-url, mp4 transcoded-url, cdn url == raw-url, vtt sub and json subs
        'url': 'https://www.loom.com/share/9458bcbf79784162aa62ffb8dd66201b',
        'md5': '7b6bfdef8181c4ffc376e18919a4dcc2',
        'info_dict': {
            'id': '9458bcbf79784162aa62ffb8dd66201b',
            'ext': 'mp4',
            'title': 'Sharing screen with gpt-4',
            'description': 'Sharing screen with GPT 4 vision model and asking questions to guide through blender.',
            'uploader': 'Suneel Matham',
            'chapters': 'count:3',
            'upload_date': '20231109',
            'timestamp': 1699518978,
            'duration': 93,
        },
    }, {
        # mpd raw-url, mp4 transcoded-url, cdn url == raw-url, no subs
        'url': 'https://www.loom.com/share/24351eb8b317420289b158e4b7e96ff2',
        'info_dict': {
            'id': '24351eb8b317420289b158e4b7e96ff2',
            'ext': 'webm',
            'title': 'OMFG clown',
            'description': 'md5:285c5ee9d62aa087b7e3271b08796815',
            'uploader': 'Brailey Bragg',
            'upload_date': '20210924',
            'timestamp': 1632519618,
            'duration': 210,
        },
        'params': {'skip_download': 'dash'},
        'expected_warnings': ['Failed to parse JSON'],  # transcoded-url no longer available
    }, {
        # password-protected
        'url': 'https://www.loom.com/share/50e26e8aeb7940189dff5630f95ce1f4',
        'md5': '5cc7655e7d55d281d203f8ffd14771f7',
        'info_dict': {
            'id': '50e26e8aeb7940189dff5630f95ce1f4',
            'ext': 'mp4',
            'title': 'iOS Mobile Upload',
            'uploader': 'Simon Curran',
            'upload_date': '20200520',
            'timestamp': 1590000123,
            'duration': 35,
        },
        'params': {'videopassword': 'seniorinfants2'},
        'expected_warnings': ['Failed to parse JSON'],  # transcoded-url no longer available
    }, {
        # embed, transcoded-url endpoint sends empty JSON response, split video and audio HLS formats
        'url': 'https://www.loom.com/embed/ddcf1c1ad21f451ea7468b1e33917e4e',
        'md5': 'f983a0f02f24331738b2f43aecb05256',
        'info_dict': {
            'id': 'ddcf1c1ad21f451ea7468b1e33917e4e',
            'ext': 'mp4',
            'title': 'CF Reset User\'s Password',
            'uploader': 'Aimee Heintz',
            'upload_date': '20220707',
            'timestamp': 1657216459,
            'duration': 181,
        },
        'params': {'format': 'bestvideo'},  # Test video-only fixup
        'expected_warnings': ['Failed to parse JSON'],
    }]
    _WEBPAGE_TESTS = [{
        'url': 'https://www.loom.com/community/e1229802a8694a09909e8ba0fbb6d073-pg',
        'md5': 'ec838cd01b576cf0386f32e1ae424609',
        'info_dict': {
            'id': 'e1229802a8694a09909e8ba0fbb6d073',
            'ext': 'mp4',
            'title': 'Rexie Jane Cimafranca - Founder\'s Presentation',
            'uploader': 'Rexie Cimafranca',
            'upload_date': '20230213',
            'duration': 247,
            'timestamp': 1676274030,
        },
        'skip': '404 Not Found',
    }]

    _GRAPHQL_VARIABLES = {
        'GetVideoSource': {
            'acceptableMimes': ['DASH', 'M3U8', 'MP4', 'WEBM'],
        },
    }
    _GRAPHQL_QUERIES = {
        'GetVideoSSR': textwrap.dedent('''\
            query GetVideoSSR($videoId: ID!, $password: String) {
              getVideo(id: $videoId, password: $password) {
                __typename
                ... on PrivateVideo {
                  id
                  status
                  message
                  __typename
                }
                ... on VideoPasswordMissingOrIncorrect {
                  id
                  message
                  __typename
                }
                ... on RegularUserVideo {
                  id
                  __typename
                  createdAt
                  description
                  download_enabled
                  folder_id
                  is_protected
                  needs_password
                  owner {
                    display_name
                    __typename
                  }
                  privacy
                  s3_id
                  name
                  video_properties {
                    avgBitRate
                    client
                    camera_enabled
                    client_version
                    duration
                    durationMs
                    format
                    height
                    microphone_enabled
                    os
                    os_version
                    recordingClient
                    recording_type
                    recording_version
                    screen_type
                    tab_audio
                    trim_duration
                    width
                    __typename
                  }
                  playable_duration
                  source_duration
                  visibility
                }
              }
            }\n'''),
        'GetVideoSource': textwrap.dedent('''\
            query GetVideoSource($videoId: ID!, $password: String, $acceptableMimes: [CloudfrontVideoAcceptableMime]) {
              getVideo(id: $videoId, password: $password) {
                ... on RegularUserVideo {
                  id
                  nullableRawCdnUrl(acceptableMimes: $acceptableMimes, password: $password) {
                    url
                    credentials {
                      Policy
                      Signature
                      KeyPairId
                      __typename
                    }
                    __typename
                  }
                  __typename
                }
                __typename
              }
            }\n'''),
        'FetchVideoTranscript': textwrap.dedent('''\
            query FetchVideoTranscript($videoId: ID!, $password: String) {
              fetchVideoTranscript(videoId: $videoId, password: $password) {
                ... on VideoTranscriptDetails {
                  id
                  video_id
                  source_url
                  captions_source_url
                  __typename
                }
                ... on GenericError {
                  message
                  __typename
                }
                __typename
              }
            }\n'''),
        'FetchChapters': textwrap.dedent('''\
            query FetchChapters($videoId: ID!, $password: String) {
              fetchVideoChapters(videoId: $videoId, password: $password) {
                ... on VideoChapters {
                  video_id
                  content
                  __typename
                }
                ... on EmptyChaptersPayload {
                  content
                  __typename
                }
                ... on InvalidRequestWarning {
                  message
                  __typename
                }
                ... on Error {
                  message
                  __typename
                }
                __typename
              }
            }\n'''),
    }
    _APOLLO_GRAPHQL_VERSION = '45a5bd4'

    def _call_graphql_api(self, operation_name, video_id, note=None, errnote=None, fatal=True):
        password = self.get_param('videopassword')
        return self._download_json(
            'https://www.loom.com/graphql', video_id, note or 'Downloading GraphQL JSON',
            errnote or 'Failed to download GraphQL JSON', headers={
                'Accept': 'application/json',
                'Content-Type': 'application/json',
                'x-loom-request-source': f'loom_web_{self._APOLLO_GRAPHQL_VERSION}',
                'apollographql-client-name': 'web',
                'apollographql-client-version': self._APOLLO_GRAPHQL_VERSION,
                'graphql-operation-name': operation_name,
                'Origin': 'https://www.loom.com',
            }, data=json.dumps({
                'operationName': operation_name,
                'variables': {
                    'videoId': video_id,
                    'password': password,
                    **self._GRAPHQL_VARIABLES.get(operation_name, {}),
                },
                'query': self._GRAPHQL_QUERIES[operation_name],
            }, separators=(',', ':')).encode(), fatal=fatal)

    def _call_url_api(self, endpoint, video_id):
        response = self._download_json(
            f'https://www.loom.com/api/campaigns/sessions/{video_id}/{endpoint}', video_id,
            f'Downloading {endpoint} JSON', f'Failed to download {endpoint} JSON', fatal=False,
            headers={'Accept': 'application/json', 'Content-Type': 'application/json'},
            data=json.dumps({
                'anonID': str(uuid.uuid4()),
                'deviceID': None,
                'force_original': False,  # HTTP error 401 if True
                'password': self.get_param('videopassword'),
            }, separators=(',', ':')).encode())
        return traverse_obj(response, ('url', {url_or_none}))

    def _extract_formats(self, video_id, metadata, video_data):
        formats = []
        video_properties = traverse_obj(metadata, ('video_properties', {
            'width': ('width', {int_or_none}),
            'height': ('height', {int_or_none}),
            'acodec': ('microphone_enabled', {lambda x: 'none' if x is False else None}),
        }))

        def get_formats(format_url, format_id, quality):
            if not format_url:
                return
            ext = determine_ext(format_url)
            query = urllib.parse.urlparse(format_url).query

            if ext == 'm3u8':
                # Extract pre-merged HLS formats to avoid buggy parsing of metadata in split playlists
                format_url = format_url.replace('-split.m3u8', '.m3u8')
                m3u8_formats = self._extract_m3u8_formats(
                    format_url, video_id, 'mp4', m3u8_id=f'hls-{format_id}', fatal=False, quality=quality)
                # Sometimes only split video/audio formats are available, need to fixup video-only formats
                is_not_premerged = 'none' in traverse_obj(m3u8_formats, (..., 'vcodec'))
                for fmt in m3u8_formats:
                    if is_not_premerged and fmt.get('vcodec') != 'none':
                        fmt['acodec'] = 'none'
                    yield {
                        **fmt,
                        'url': update_url(fmt['url'], query=query),
                        'extra_param_to_segment_url': query,
                    }

            elif ext == 'mpd':
                dash_formats = self._extract_mpd_formats(
                    format_url, video_id, mpd_id=f'dash-{format_id}', fatal=False)
                for fmt in dash_formats:
                    yield {
                        **fmt,
                        'extra_param_to_segment_url': query,
                        'quality': quality,
                    }

            else:
                yield {
                    'url': format_url,
                    'ext': ext,
                    'format_id': f'http-{format_id}',
                    'quality': quality,
                    **video_properties,
                }

        raw_url = self._call_url_api('raw-url', video_id)
        formats.extend(get_formats(raw_url, 'raw', quality=1))  # original quality

        transcoded_url = self._call_url_api('transcoded-url', video_id)
        formats.extend(get_formats(transcoded_url, 'transcoded', quality=-1))  # transcoded quality

        cdn_url = traverse_obj(video_data, ('data', 'getVideo', 'nullableRawCdnUrl', 'url', {url_or_none}))
        # cdn_url is usually a dupe, but the raw-url/transcoded-url endpoints could return errors
        valid_urls = [update_url(url, query=None) for url in (raw_url, transcoded_url) if url]
        if cdn_url and update_url(cdn_url, query=None) not in valid_urls:
            formats.extend(get_formats(cdn_url, 'cdn', quality=0))  # could be original or transcoded

        return formats

    def _get_subtitles(self, video_id):
        subs_data = self._call_graphql_api(
            'FetchVideoTranscript', video_id, 'Downloading GraphQL subtitles JSON', fatal=False)
        return filter_dict({
            'en': traverse_obj(subs_data, (
                'data', 'fetchVideoTranscript',
                ('source_url', 'captions_source_url'), {
                    'url': {url_or_none},
                })) or None,
        })

    def _real_extract(self, url):
        video_id = self._match_id(url)
        metadata = traverse_obj(
            self._call_graphql_api('GetVideoSSR', video_id, 'Downloading GraphQL metadata JSON', fatal=False),
            ('data', 'getVideo', {dict})) or {}

        if metadata.get('__typename') == 'VideoPasswordMissingOrIncorrect':
            if not self.get_param('videopassword'):
                raise ExtractorError(
                    'This video is password-protected, use the --video-password option', expected=True)
            raise ExtractorError('Invalid video password', expected=True)

        video_data = self._call_graphql_api(
            'GetVideoSource', video_id, 'Downloading GraphQL video JSON')
        chapter_data = self._call_graphql_api(
            'FetchChapters', video_id, 'Downloading GraphQL chapters JSON', fatal=False)
        duration = traverse_obj(metadata, ('video_properties', 'duration', {int_or_none}))

        return {
            'id': video_id,
            'duration': duration,
            'chapters': self._extract_chapters_from_description(
                traverse_obj(chapter_data, ('data', 'fetchVideoChapters', 'content', {str})), duration) or None,
            'formats': self._extract_formats(video_id, metadata, video_data),
            'subtitles': self.extract_subtitles(video_id),
            **traverse_obj(metadata, {
                'title': ('name', {str}),
                'description': ('description', {str}),
                'uploader': ('owner', 'display_name', {str}),
                'timestamp': ('createdAt', {parse_iso8601}),
            }),
        }


class LoomFolderIE(InfoExtractor):
    _WORKING = False
    IE_NAME = 'loom:folder'
    _VALID_URL = r'https?://(?:www\.)?loom\.com/share/folder/(?P<id>[\da-f]{32})'
    _TESTS = [{
        # 2 subfolders, no videos in root
        'url': 'https://www.loom.com/share/folder/997db4db046f43e5912f10dc5f817b5c',
        'playlist_mincount': 16,
        'info_dict': {
            'id': '997db4db046f43e5912f10dc5f817b5c',
            'title': 'Blending Lessons',
        },
    }, {
        # only videos, no subfolders
        'url': 'https://www.loom.com/share/folder/9a8a87f6b6f546d9a400c8e7575ff7f2',
        'playlist_mincount': 12,
        'info_dict': {
            'id': '9a8a87f6b6f546d9a400c8e7575ff7f2',
            'title': 'List A- a, i, o',
        },
    }, {
        # videos in root and empty subfolder
        'url': 'https://www.loom.com/share/folder/886e534218c24fd292e97e9563078cc4',
        'playlist_mincount': 21,
        'info_dict': {
            'id': '886e534218c24fd292e97e9563078cc4',
            'title': 'Medicare Agent Training videos',
        },
    }, {
        # videos in root and videos in subfolders
        'url': 'https://www.loom.com/share/folder/b72c4ecdf04745da9403926d80a40c38',
        'playlist_mincount': 21,
        'info_dict': {
            'id': 'b72c4ecdf04745da9403926d80a40c38',
            'title': 'Quick Altos Q & A Tutorials',
        },
    }, {
        # recursive folder extraction
        'url': 'https://www.loom.com/share/folder/8b458a94e0e4449b8df9ea7a68fafc4e',
        'playlist_count': 23,
        'info_dict': {
            'id': '8b458a94e0e4449b8df9ea7a68fafc4e',
            'title': 'Sezer Texting Guide',
        },
    }, {
        # more than 50 videos in 1 folder
        'url': 'https://www.loom.com/share/folder/e056a91d290d47ca9b00c9d1df56c463',
        'playlist_mincount': 61,
        'info_dict': {
            'id': 'e056a91d290d47ca9b00c9d1df56c463',
            'title': 'User Videos',
        },
    }, {
        # many subfolders
        'url': 'https://www.loom.com/share/folder/c2dde8cc67454f0e99031677279d8954',
        'playlist_mincount': 75,
        'info_dict': {
            'id': 'c2dde8cc67454f0e99031677279d8954',
            'title': 'Honors 1',
        },
    }, {
        'url': 'https://www.loom.com/share/folder/bae17109a68146c7803454f2893c8cf8/Edpuzzle',
        'only_matching': True,
    }]

    def _extract_folder_data(self, folder_id):
        return self._download_json(
            f'https://www.loom.com/v1/folders/{folder_id}', folder_id,
            'Downloading folder info JSON', query={'limit': '10000'})

    def _extract_folder_entries(self, folder_id, initial_folder_data=None):
        folder_data = initial_folder_data or self._extract_folder_data(folder_id)

        for video in traverse_obj(folder_data, ('videos', lambda _, v: v['id'])):
            video_id = video['id']
            yield self.url_result(
                f'https://www.loom.com/share/{video_id}', LoomIE, video_id, video.get('name'))

        # Recurse into subfolders
        for subfolder_id in traverse_obj(folder_data, (
                'folders', lambda _, v: v['id'] != folder_id, 'id', {str})):
            yield from self._extract_folder_entries(subfolder_id)

    def _real_extract(self, url):
        playlist_id = self._match_id(url)
        playlist_data = self._extract_folder_data(playlist_id)

        return self.playlist_result(
            self._extract_folder_entries(playlist_id, playlist_data), playlist_id,
            traverse_obj(playlist_data, ('folder', 'name', {str.strip})))



from .common import InfoExtractor
from ..utils import (
    ExtractorError,
    float_or_none,
    format_field,
    int_or_none,
    parse_iso8601,
    parse_qs,
    truncate_string,
    url_or_none,
)
from ..utils.traversal import traverse_obj, value


class BandlabBaseIE(InfoExtractor):
    def _call_api(self, endpoint, asset_id, **kwargs):
        headers = kwargs.pop('headers', None) or {}
        return self._download_json(
            f'https://www.bandlab.com/api/v1.3/{endpoint}/{asset_id}',
            asset_id, headers={
                'accept': 'application/json',
                'referer': 'https://www.bandlab.com/',
                'x-client-id': 'BandLab-Web',
                'x-client-version': '10.1.124',
                **headers,
            }, **kwargs)

    def _parse_revision(self, revision_data, url=None):
        return {
            'vcodec': 'none',
            'media_type': 'revision',
            'extractor_key': BandlabIE.ie_key(),
            'extractor': BandlabIE.IE_NAME,
            **traverse_obj(revision_data, {
                'webpage_url': (
                    'id', ({value(url)}, {format_field(template='https://www.bandlab.com/revision/%s')}), filter, any),
                'id': (('revisionId', 'id'), {str}, any),
                'title': ('song', 'name', {str}),
                'track': ('song', 'name', {str}),
                'url': ('mixdown', 'file', {url_or_none}),
                'thumbnail': ('song', 'picture', 'url', {url_or_none}),
                'description': ('description', {str}),
                'uploader': ('creator', 'name', {str}),
                'uploader_id': ('creator', 'username', {str}),
                'timestamp': ('createdOn', {parse_iso8601}),
                'duration': ('mixdown', 'duration', {float_or_none}),
                'view_count': ('counters', 'plays', {int_or_none}),
                'like_count': ('counters', 'likes', {int_or_none}),
                'comment_count': ('counters', 'comments', {int_or_none}),
                'genres': ('genres', ..., 'name', {str}),
            }),
        }

    def _parse_track(self, track_data, url=None):
        return {
            'vcodec': 'none',
            'media_type': 'track',
            'extractor_key': BandlabIE.ie_key(),
            'extractor': BandlabIE.IE_NAME,
            **traverse_obj(track_data, {
                'webpage_url': (
                    'id', ({value(url)}, {format_field(template='https://www.bandlab.com/post/%s')}), filter, any),
                'id': (('revisionId', 'id'), {str}, any),
                'url': ('track', 'sample', 'audioUrl', {url_or_none}),
                'title': ('track', 'name', {str}),
                'track': ('track', 'name', {str}),
                'description': ('caption', {str}),
                'thumbnail': ('track', 'picture', ('original', 'url'), {url_or_none}, any),
                'view_count': ('counters', 'plays', {int_or_none}),
                'like_count': ('counters', 'likes', {int_or_none}),
                'comment_count': ('counters', 'comments', {int_or_none}),
                'duration': ('track', 'sample', 'duration', {float_or_none}),
                'uploader': ('creator', 'name', {str}),
                'uploader_id': ('creator', 'username', {str}),
                'timestamp': ('createdOn', {parse_iso8601}),
            }),
        }

    def _parse_video(self, video_data, url=None):
        return {
            'media_type': 'video',
            'extractor_key': BandlabIE.ie_key(),
            'extractor': BandlabIE.IE_NAME,
            **traverse_obj(video_data, {
                'id': ('id', {str}),
                'webpage_url': (
                    'id', ({value(url)}, {format_field(template='https://www.bandlab.com/post/%s')}), filter, any),
                'url': ('video', 'url', {url_or_none}),
                'title': ('caption', {lambda x: x.replace('\n', ' ')}, {truncate_string(left=72)}),
                'description': ('caption', {str}),
                'thumbnail': ('video', 'picture', 'url', {url_or_none}),
                'view_count': ('video', 'counters', 'plays', {int_or_none}),
                'like_count': ('video', 'counters', 'likes', {int_or_none}),
                'comment_count': ('counters', 'comments', {int_or_none}),
                'duration': ('video', 'duration', {float_or_none}),
                'uploader': ('creator', 'name', {str}),
                'uploader_id': ('creator', 'username', {str}),
            }),
        }


class BandlabIE(BandlabBaseIE):
    _VALID_URL = [
        r'https?://(?:www\.)?bandlab.com/(?P<url_type>track|post|revision)/(?P<id>[\da-f_-]+)',
        r'https?://(?:www\.)?bandlab.com/(?P<url_type>embed)/\?(?:[^#]*&)?id=(?P<id>[\da-f-]+)',
    ]
    _EMBED_REGEX = [rf'<iframe[^>]+src=[\'"](?P<url>{_VALID_URL[1]})[\'"]']
    _TESTS = [{
        'url': 'https://www.bandlab.com/track/04b37e88dba24967b9dac8eb8567ff39_07d7f906fc96ee11b75e000d3a428fff',
        'md5': '46f7b43367dd268bbcf0bbe466753b2c',
        'info_dict': {
            'id': '02d7f906-fc96-ee11-b75e-000d3a428fff',
            'ext': 'm4a',
            'uploader_id': 'ender_milze',
            'track': 'sweet black',
            'description': 'composed by juanjn3737',
            'timestamp': 1702171963,
            'view_count': int,
            'like_count': int,
            'duration': 54.629999999999995,
            'title': 'sweet black',
            'upload_date': '20231210',
            'thumbnail': 'https://bl-prod-images.azureedge.net/v1.0/songs/fa082beb-b856-4730-9170-a57e4e32cc2c/',
            'genres': ['Lofi'],
            'uploader': 'ender milze',
            'comment_count': int,
            'media_type': 'revision',
        },
    }, {
        # Same track as above but post URL
        'url': 'https://www.bandlab.com/post/07d7f906-fc96-ee11-b75e-000d3a428fff',
        'md5': '46f7b43367dd268bbcf0bbe466753b2c',
        'info_dict': {
            'id': '02d7f906-fc96-ee11-b75e-000d3a428fff',
            'ext': 'm4a',
            'uploader_id': 'ender_milze',
            'track': 'sweet black',
            'description': 'composed by juanjn3737',
            'timestamp': 1702171973,
            'view_count': int,
            'like_count': int,
            'duration': 54.629999999999995,
            'title': 'sweet black',
            'upload_date': '20231210',
            'thumbnail': 'https://bl-prod-images.azureedge.net/v1.0/songs/fa082beb-b856-4730-9170-a57e4e32cc2c/',
            'genres': ['Lofi'],
            'uploader': 'ender milze',
            'comment_count': int,
            'media_type': 'revision',
        },
    }, {
        # SharedKey Example
        'url': 'https://www.bandlab.com/track/048916c2-c6da-ee11-85f9-6045bd2e11f9?sharedKey=0NNWX8qYAEmI38lWAzCNDA',
        'md5': '15174b57c44440e2a2008be9cae00250',
        'info_dict': {
            'id': '038916c2-c6da-ee11-85f9-6045bd2e11f9',
            'ext': 'm4a',
            'comment_count': int,
            'genres': ['Other'],
            'uploader_id': 'user8353034818103753',
            'thumbnail': 'https://bl-prod-images.azureedge.net/v1.0/songs/51b18363-da23-4b9b-a29c-2933a3e561ca/',
            'timestamp': 1709625771,
            'track': 'PodcastMaerchen4b',
            'duration': 468.14,
            'view_count': int,
            'description': 'Podcast: Neues aus der Märchenwelt',
            'like_count': int,
            'upload_date': '20240305',
            'uploader': 'Erna Wageneder',
            'title': 'PodcastMaerchen4b',
            'media_type': 'revision',
        },
    }, {
        # Different Revision selected
        'url': 'https://www.bandlab.com/track/130343fc-148b-ea11-96d2-0003ffd1fc09?revId=110343fc-148b-ea11-96d2-0003ffd1fc09',
        'md5': '74e055ef9325d63f37088772fbfe4454',
        'info_dict': {
            'id': '110343fc-148b-ea11-96d2-0003ffd1fc09',
            'ext': 'm4a',
            'timestamp': 1588273294,
            'thumbnail': 'https://bl-prod-images.azureedge.net/v1.0/users/b612e533-e4f7-4542-9f50-3fcfd8dd822c/',
            'description': 'Final Revision.',
            'title': 'Replay ( Instrumental)',
            'uploader': 'David R Sparks',
            'uploader_id': 'davesnothome69',
            'view_count': int,
            'comment_count': int,
            'track': 'Replay ( Instrumental)',
            'genres': ['Rock'],
            'upload_date': '20200430',
            'like_count': int,
            'duration': 279.43,
            'media_type': 'revision',
        },
    }, {
        # Video
        'url': 'https://www.bandlab.com/post/5cdf9036-3857-ef11-991a-6045bd36e0d9',
        'md5': '8caa2ef28e86c1dacf167293cfdbeba9',
        'info_dict': {
            'id': '5cdf9036-3857-ef11-991a-6045bd36e0d9',
            'ext': 'mp4',
            'duration': 44.705,
            'thumbnail': 'https://bl-prod-images.azureedge.net/v1.0/videos/67c6cef1-cef6-40d3-831e-a55bc1dcb972/',
            'comment_count': int,
            'title': 'backing vocals',
            'uploader_id': 'marliashya',
            'uploader': 'auraa',
            'like_count': int,
            'description': 'backing vocals',
            'media_type': 'video',
        },
    }, {
        # Embed Example
        'url': 'https://www.bandlab.com/embed/?blur=false&id=014de0a4-7d82-ea11-a94c-0003ffd19c0f',
        'md5': 'a4ad05cb68c54faaed9b0a8453a8cf4a',
        'info_dict': {
            'id': '014de0a4-7d82-ea11-a94c-0003ffd19c0f',
            'ext': 'm4a',
            'comment_count': int,
            'genres': ['Electronic'],
            'uploader': 'Charlie Henson',
            'timestamp': 1587328674,
            'upload_date': '20200419',
            'view_count': int,
            'track': 'Positronic Meltdown',
            'duration': 318.55,
            'thumbnail': 'https://bl-prod-images.azureedge.net/v1.0/songs/87165bc3-5439-496e-b1f7-a9f13b541ff2/',
            'description': 'Checkout my tracks at AOMX http://aomxsounds.com/',
            'uploader_id': 'microfreaks',
            'title': 'Positronic Meltdown',
            'like_count': int,
            'media_type': 'revision',
        },
    }, {
        # Track without revisions available
        'url': 'https://www.bandlab.com/track/55767ac51789ea11a94c0003ffd1fc09_2f007b0a37b94ec7a69bc25ae15108a5',
        'md5': 'f05d68a3769952c2d9257c473e14c15f',
        'info_dict': {
            'id': '55767ac51789ea11a94c0003ffd1fc09_2f007b0a37b94ec7a69bc25ae15108a5',
            'ext': 'm4a',
            'track': 'insame',
            'like_count': int,
            'duration': 84.03,
            'title': 'insame',
            'view_count': int,
            'comment_count': int,
            'uploader': 'Sorakime',
            'uploader_id': 'sorakime',
            'thumbnail': 'https://bl-prod-images.azureedge.net/v1.0/users/572a351a-0f3a-4c6a-ac39-1a5defdeeb1c/',
            'timestamp': 1691162128,
            'upload_date': '20230804',
            'media_type': 'track',
        },
    }, {
        'url': 'https://www.bandlab.com/revision/014de0a4-7d82-ea11-a94c-0003ffd19c0f',
        'only_matching': True,
    }]
    _WEBPAGE_TESTS = [{
        'url': 'https://phantomluigi.github.io/',
        'info_dict': {
            'id': 'e14223c3-7871-ef11-bdfd-000d3a980db3',
            'ext': 'm4a',
            'view_count': int,
            'upload_date': '20240913',
            'uploader_id': 'phantommusicofficial',
            'timestamp': 1726194897,
            'uploader': 'Phantom',
            'comment_count': int,
            'genres': ['Progresive Rock'],
            'description': 'md5:a38cd668f7a2843295ef284114f18429',
            'duration': 225.23,
            'like_count': int,
            'title': 'Vermilion Pt. 2 (Cover)',
            'track': 'Vermilion Pt. 2 (Cover)',
            'thumbnail': 'https://bandlabimages.azureedge.net/v1.0/songs/62b10750-7aef-4f42-ad08-1af52f577e97/',
            'media_type': 'revision',
        },
    }]

    def _real_extract(self, url):
        display_id, url_type = self._match_valid_url(url).group('id', 'url_type')

        qs = parse_qs(url)
        revision_id = traverse_obj(qs, (('revId', 'id'), 0, any))
        if url_type == 'revision':
            revision_id = display_id

        revision_data = None
        if not revision_id:
            post_data = self._call_api(
                'posts', display_id, note='Downloading post data',
                query=traverse_obj(qs, {'sharedKey': ('sharedKey', 0)}))

            revision_id = traverse_obj(post_data, (('revisionId', ('revision', 'id')), {str}, any))
            revision_data = traverse_obj(post_data, ('revision', {dict}))

            if not revision_data and not revision_id:
                post_type = post_data.get('type')
                if post_type == 'Video':
                    return self._parse_video(post_data, url=url)
                if post_type == 'Track':
                    return self._parse_track(post_data, url=url)
                raise ExtractorError(f'Could not extract data for post type {post_type!r}')

        if not revision_data:
            revision_data = self._call_api(
                'revisions', revision_id, note='Downloading revision data', query={'edit': 'false'})

        return self._parse_revision(revision_data, url=url)


class BandlabPlaylistIE(BandlabBaseIE):
    _VALID_URL = [
        r'https?://(?:www\.)?bandlab.com/(?:[\w]+/)?(?P<type>albums|collections)/(?P<id>[\da-f-]+)',
        r'https?://(?:www\.)?bandlab.com/(?P<type>embed)/collection/\?(?:[^#]*&)?id=(?P<id>[\da-f-]+)',
    ]
    _EMBED_REGEX = [rf'<iframe[^>]+src=[\'"](?P<url>{_VALID_URL[1]})[\'"]']
    _TESTS = [{
        'url': 'https://www.bandlab.com/davesnothome69/albums/89b79ea6-de42-ed11-b495-00224845aac7',
        'info_dict': {
            'thumbnail': 'https://bl-prod-images.azureedge.net/v1.3/albums/69507ff3-579a-45be-afca-9e87eddec944/',
            'release_date': '20221003',
            'title': 'Remnants',
            'album': 'Remnants',
            'like_count': int,
            'album_type': 'LP',
            'description': 'A collection of some feel good, rock hits.',
            'comment_count': int,
            'view_count': int,
            'id': '89b79ea6-de42-ed11-b495-00224845aac7',
            'uploader': 'David R Sparks',
            'uploader_id': 'davesnothome69',
        },
        'playlist_count': 10,
    }, {
        'url': 'https://www.bandlab.com/slytheband/collections/955102d4-1040-ef11-86c3-000d3a42581b',
        'info_dict': {
            'id': '955102d4-1040-ef11-86c3-000d3a42581b',
            'timestamp': 1720762659,
            'view_count': int,
            'title': 'My Shit 🖤',
            'uploader_id': 'slytheband',
            'uploader': '𝓢𝓛𝓨',
            'upload_date': '20240712',
            'like_count': int,
            'thumbnail': 'https://bandlabimages.azureedge.net/v1.0/collections/2c64ca12-b180-4b76-8587-7a8da76bddc8/',
        },
        'playlist_count': 15,
    }, {
        # Embeds can contain both albums and collections with the same URL pattern. This is an album
        'url': 'https://www.bandlab.com/embed/collection/?id=12cc6f7f-951b-ee11-907c-00224844f303',
        'info_dict': {
            'id': '12cc6f7f-951b-ee11-907c-00224844f303',
            'release_date': '20230706',
            'description': 'This is a collection of songs I created when I had an Amiga computer.',
            'view_count': int,
            'title': 'Mark Salud The Amiga Collection',
            'uploader_id': 'mssirmooth1962',
            'comment_count': int,
            'thumbnail': 'https://bl-prod-images.azureedge.net/v1.3/albums/d618bd7b-0537-40d5-bdd8-61b066e77d59/',
            'like_count': int,
            'uploader': 'Mark Salud',
            'album': 'Mark Salud The Amiga Collection',
            'album_type': 'LP',
        },
        'playlist_count': 24,
    }, {
        # Tracks without revision id
        'url': 'https://www.bandlab.com/embed/collection/?id=e98aafb5-d932-ee11-b8f0-00224844c719',
        'info_dict': {
            'like_count': int,
            'uploader_id': 'sorakime',
            'comment_count': int,
            'uploader': 'Sorakime',
            'view_count': int,
            'description': 'md5:4ec31c568a5f5a5a2b17572ea64c3825',
            'release_date': '20230812',
            'title': 'Art',
            'album': 'Art',
            'album_type': 'Album',
            'id': 'e98aafb5-d932-ee11-b8f0-00224844c719',
            'thumbnail': 'https://bl-prod-images.azureedge.net/v1.3/albums/20c890de-e94a-4422-828a-2da6377a13c8/',
        },
        'playlist_count': 13,
    }, {
        'url': 'https://www.bandlab.com/albums/89b79ea6-de42-ed11-b495-00224845aac7',
        'only_matching': True,
    }]

    def _entries(self, album_data):
        for post in traverse_obj(album_data, ('posts', lambda _, v: v['type'])):
            post_type = post['type']
            if post_type == 'Revision':
                yield self._parse_revision(post.get('revision'))
            elif post_type == 'Track':
                yield self._parse_track(post)
            elif post_type == 'Video':
                yield self._parse_video(post)
            else:
                self.report_warning(f'Skipping unknown post type: "{post_type}"')

    def _real_extract(self, url):
        playlist_id, playlist_type = self._match_valid_url(url).group('id', 'type')

        endpoints = {
            'albums': ['albums'],
            'collections': ['collections'],
            'embed': ['collections', 'albums'],
        }.get(playlist_type)
        for endpoint in endpoints:
            playlist_data = self._call_api(
                endpoint, playlist_id, note=f'Downloading {endpoint[:-1]} data',
                fatal=False, expected_status=404)
            if not playlist_data.get('errorCode'):
                playlist_type = endpoint
                break
        if error_code := playlist_data.get('errorCode'):
            raise ExtractorError(f'Could not find playlist data. Error code: "{error_code}"')

        return self.playlist_result(
            self._entries(playlist_data), playlist_id,
            **traverse_obj(playlist_data, {
                'title': ('name', {str}),
                'description': ('description', {str}),
                'uploader': ('creator', 'name', {str}),
                'uploader_id': ('creator', 'username', {str}),
                'timestamp': ('createdOn', {parse_iso8601}),
                'release_date': ('releaseDate', {lambda x: x.replace('-', '')}, filter),
                'thumbnail': ('picture', ('original', 'url'), {url_or_none}, any),
                'like_count': ('counters', 'likes', {int_or_none}),
                'comment_count': ('counters', 'comments', {int_or_none}),
                'view_count': ('counters', 'plays', {int_or_none}),
            }),
            **(traverse_obj(playlist_data, {
                'album': ('name', {str}),
                'album_type': ('type', {str}),
            }) if playlist_type == 'albums' else {}))



import re

from .common import InfoExtractor
from ..utils import (
    NO_DEFAULT,
    int_or_none,
    parse_duration,
    str_to_int,
)


class DrTuberIE(InfoExtractor):
    _VALID_URL = r'https?://(?:(?:www|m)\.)?drtuber\.com/(?:video|embed)/(?P<id>\d+)(?:/(?P<display_id>[\w-]+))?'
    _EMBED_REGEX = [r'<iframe[^>]+?src=["\'](?P<url>(?:https?:)?//(?:www\.)?drtuber\.com/embed/\d+)']
    _TESTS = [{
        'url': 'http://www.drtuber.com/video/1740434/hot-perky-blonde-naked-golf',
        'md5': '93e680cf2536ad0dfb7e74d94a89facd',
        'info_dict': {
            'id': '1740434',
            'display_id': 'hot-perky-blonde-naked-golf',
            'ext': 'mp4',
            'title': 'hot perky blonde naked golf',
            'like_count': int,
            'comment_count': int,
            'categories': ['Babe', 'Blonde', 'Erotic', 'Outdoor', 'Softcore', 'Solo'],
            'thumbnail': r're:https?://.*\.jpg$',
            'age_limit': 18,
        },
    }, {
        'url': 'http://www.drtuber.com/embed/489939',
        'only_matching': True,
    }, {
        'url': 'http://m.drtuber.com/video/3893529/lingerie-blowjob-from-beautiful-teen',
        'only_matching': True,
    }]

    def _real_extract(self, url):
        mobj = self._match_valid_url(url)
        video_id = mobj.group('id')
        display_id = mobj.group('display_id') or video_id

        webpage = self._download_webpage(
            f'http://www.drtuber.com/video/{video_id}', display_id)

        video_data = self._download_json(
            'http://www.drtuber.com/player_config_json/', video_id, query={
                'vid': video_id,
                'embed': 0,
                'aid': 0,
                'domain_id': 0,
            })

        formats = []
        for format_id, video_url in video_data['files'].items():
            if video_url:
                formats.append({
                    'format_id': format_id,
                    'quality': 2 if format_id == 'hq' else 1,
                    'url': video_url,
                })

        duration = int_or_none(video_data.get('duration')) or parse_duration(
            video_data.get('duration_format'))

        title = self._html_search_regex(
            (r'<h1[^>]+class=["\']title[^>]+>([^<]+)',
             r'<title>([^<]+)\s*@\s+DrTuber',
             r'class="title_watch"[^>]*><(?:p|h\d+)[^>]*>([^<]+)<',
             r'<p[^>]+class="title_substrate">([^<]+)</p>',
             r'<title>([^<]+) - \d+'),
            webpage, 'title')

        thumbnail = self._html_search_regex(
            r'poster="([^"]+)"',
            webpage, 'thumbnail', fatal=False)

        def extract_count(id_, name, default=NO_DEFAULT):
            return str_to_int(self._html_search_regex(
                rf'<span[^>]+(?:class|id)="{id_}"[^>]*>([\d,\.]+)</span>',
                webpage, f'{name} count', default=default, fatal=False))

        like_count = extract_count('rate_likes', 'like')
        dislike_count = extract_count('rate_dislikes', 'dislike', default=None)
        comment_count = extract_count('comments_count', 'comment')

        cats_str = self._search_regex(
            r'<div[^>]+class="categories_list">(.+?)</div>',
            webpage, 'categories', fatal=False)
        categories = [] if not cats_str else re.findall(
            r'<a title="([^"]+)"', cats_str)

        return {
            'id': video_id,
            'display_id': display_id,
            'formats': formats,
            'title': title,
            'thumbnail': thumbnail,
            'like_count': like_count,
            'dislike_count': dislike_count,
            'comment_count': comment_count,
            'categories': categories,
            'age_limit': self._rta_search(webpage),
            'duration': duration,
        }



import json

from .common import InfoExtractor
from .vimeo import VimeoIE
from ..utils import (
    clean_html,
    extract_attributes,
    get_element_html_by_id,
    int_or_none,
    parse_duration,
    str_or_none,
    unified_strdate,
    url_or_none,
    urljoin,
)
from ..utils.traversal import traverse_obj


class LaracastsBaseIE(InfoExtractor):
    def _get_prop_data(self, url, display_id):
        webpage = self._download_webpage(url, display_id)
        return traverse_obj(
            get_element_html_by_id('app', webpage),
            ({extract_attributes}, 'data-page', {json.loads}, 'props'))

    def _parse_episode(self, episode):
        if not traverse_obj(episode, 'vimeoId'):
            self.raise_login_required('This video is only available for subscribers.')
        return self.url_result(
            VimeoIE._smuggle_referrer(
                f'https://player.vimeo.com/video/{episode["vimeoId"]}', 'https://laracasts.com/'),
            VimeoIE, url_transparent=True,
            **traverse_obj(episode, {
                'id': ('id', {int}, {str_or_none}),
                'webpage_url': ('path', {urljoin('https://laracasts.com')}),
                'title': ('title', {clean_html}),
                'season_number': ('chapter', {int_or_none}),
                'episode_number': ('position', {int_or_none}),
                'description': ('body', {clean_html}),
                'thumbnail': ('largeThumbnail', {url_or_none}),
                'duration': ('length', {int_or_none}),
                'upload_date': ('dateSegments', 'published', {unified_strdate}),
            }))


class LaracastsIE(LaracastsBaseIE):
    IE_NAME = 'laracasts'
    _VALID_URL = r'https?://(?:www\.)?laracasts\.com/series/(?P<id>[\w-]+/episodes/\d+)/?(?:[?#]|$)'
    _TESTS = [{
        'url': 'https://laracasts.com/series/30-days-to-learn-laravel-11/episodes/1',
        'md5': 'c8f5e7b02ad0e438ef9280a08c8493dc',
        'info_dict': {
            'id': '922040563',
            'title': 'Hello, Laravel',
            'ext': 'mp4',
            'duration': 519,
            'upload_date': '20240312',
            'thumbnail': 'https://laracasts.s3.amazonaws.com/videos/thumbnails/youtube/30-days-to-learn-laravel-11-1.png',
            'description': 'md5:ddd658bb241975871d236555657e1dd1',
            'season_number': 1,
            'season': 'Season 1',
            'episode_number': 1,
            'episode': 'Episode 1',
            'uploader': 'Laracasts',
            'uploader_id': 'user20182673',
            'uploader_url': 'https://vimeo.com/user20182673',
        },
        'expected_warnings': ['Failed to parse XML'],  # TODO: Remove when vimeo extractor is fixed
    }]

    def _real_extract(self, url):
        display_id = self._match_id(url)
        return self._parse_episode(self._get_prop_data(url, display_id)['lesson'])


class LaracastsPlaylistIE(LaracastsBaseIE):
    IE_NAME = 'laracasts:series'
    _VALID_URL = r'https?://(?:www\.)?laracasts\.com/series/(?P<id>[\w-]+)/?(?:[?#]|$)'
    _TESTS = [{
        'url': 'https://laracasts.com/series/30-days-to-learn-laravel-11',
        'info_dict': {
            'title': '30 Days to Learn Laravel',
            'id': '210',
            'thumbnail': 'https://laracasts.s3.amazonaws.com/series/thumbnails/social-cards/30-days-to-learn-laravel-11.png?v=2',
            'duration': 30600.0,
            'modified_date': '20240511',
            'description': 'md5:27c260a1668a450984e8f901579912dd',
            'categories': ['Frameworks'],
            'tags': ['Laravel'],
            'display_id': '30-days-to-learn-laravel-11',
        },
        'playlist_count': 30,
    }]

    def _real_extract(self, url):
        display_id = self._match_id(url)
        series = self._get_prop_data(url, display_id)['series']

        metadata = {
            'display_id': display_id,
            **traverse_obj(series, {
                'title': ('title', {str}),
                'id': ('id', {int}, {str_or_none}),
                'description': ('body', {clean_html}),
                'thumbnail': (('large_thumbnail', 'thumbnail'), {url_or_none}, any),
                'duration': ('runTime', {parse_duration}),
                'categories': ('taxonomy', 'name', {str}, all, filter),
                'tags': ('topics', ..., 'name', {str}),
                'modified_date': ('lastUpdated', {unified_strdate}),
            }),
        }

        return self.playlist_result(traverse_obj(
            series, ('chapters', ..., 'episodes', lambda _, v: v['vimeoId'], {self._parse_episode})), **metadata)



from .francetv import FranceTVBaseInfoExtractor


class LumniIE(FranceTVBaseInfoExtractor):
    _VALID_URL = r'https?://(?:www\.)?lumni\.fr/video/(?P<id>[\w-]+)'
    _TESTS = [{
        'url': 'https://www.lumni.fr/video/l-homme-et-son-environnement-dans-la-revolution-industrielle',
        'md5': '960e8240c4f2c7a20854503a71e52f5e',
        'info_dict': {
            'id': 'd2b9a4e5-a526-495b-866c-ab72737e3645',
            'ext': 'mp4',
            'title': "L'homme et son environnement dans la révolution industrielle - L'ère de l'homme",
            'thumbnail': 'https://assets.webservices.francetelevisions.fr/v1/assets/images/a7/17/9f/a7179f5f-63a5-4e11-8d4d-012ab942d905.jpg',
            'duration': 230,
        },
    }]

    def _real_extract(self, url):
        display_id = self._match_id(url)
        webpage = self._download_webpage(url, display_id)
        video_id = self._html_search_regex(
            r'<div[^>]+data-factoryid\s*=\s*["\']([^"\']+)', webpage, 'video id')
        return self._make_url_result(video_id, url=url)



import io
import quopri
import re
import uuid

from .fragment import FragmentFD
from ..compat import imghdr
from ..utils import escapeHTML, formatSeconds, srt_subtitles_timecode, urljoin
from ..version import __version__ as YT_DLP_VERSION


class MhtmlFD(FragmentFD):
    _STYLESHEET = '''\
html, body {
    margin: 0;
    padding: 0;
    height: 100vh;
}

html {
    overflow-y: scroll;
    scroll-snap-type: y mandatory;
}

body {
    scroll-snap-type: y mandatory;
    display: flex;
    flex-flow: column;
}

body > figure {
    max-width: 100vw;
    max-height: 100vh;
    scroll-snap-align: center;
}

body > figure > figcaption {
    text-align: center;
    height: 2.5em;
}

body > figure > img {
    display: block;
    margin: auto;
    max-width: 100%;
    max-height: calc(100vh - 5em);
}
'''
    _STYLESHEET = re.sub(r'\s+', ' ', _STYLESHEET)
    _STYLESHEET = re.sub(r'\B \B|(?<=[\w\-]) (?=[^\w\-])|(?<=[^\w\-]) (?=[\w\-])', '', _STYLESHEET)

    @staticmethod
    def _escape_mime(s):
        return '=?utf-8?Q?' + (b''.join(
            bytes((b,)) if b >= 0x20 else b'=%02X' % b
            for b in quopri.encodestring(s.encode(), header=True)
        )).decode('us-ascii') + '?='

    def _gen_cid(self, i, fragment, frag_boundary):
        return f'{i}.{frag_boundary}@yt-dlp.github.io.invalid'

    def _gen_stub(self, *, fragments, frag_boundary, title):
        output = io.StringIO()

        output.write(
            '<!DOCTYPE html>'
            '<html>'
            '<head>'
            f'<meta name="generator" content="yt-dlp {escapeHTML(YT_DLP_VERSION)}">'
            f'<title>{escapeHTML(title)}</title>'
            f'<style>{self._STYLESHEET}</style>'
            '<body>')

        t0 = 0
        for i, frag in enumerate(fragments):
            output.write('<figure>')
            try:
                t1 = t0 + frag['duration']
                output.write((
                    '<figcaption>Slide #{num}: {t0} – {t1} (duration: {duration})</figcaption>'
                ).format(
                    num=i + 1,
                    t0=srt_subtitles_timecode(t0),
                    t1=srt_subtitles_timecode(t1),
                    duration=formatSeconds(frag['duration'], msec=True),
                ))
            except (KeyError, ValueError, TypeError):
                t1 = None
                output.write(f'<figcaption>Slide #{i + 1}</figcaption>')
            output.write(f'<img src="cid:{self._gen_cid(i, frag, frag_boundary)}">')
            output.write('</figure>')
            t0 = t1

        return output.getvalue()

    def real_download(self, filename, info_dict):
        fragment_base_url = info_dict.get('fragment_base_url')
        fragments = info_dict['fragments'][:1] if self.params.get(
            'test', False) else info_dict['fragments']
        title = info_dict.get('title', info_dict['format_id'])
        origin = info_dict.get('webpage_url', info_dict['url'])

        ctx = {
            'filename': filename,
            'total_frags': len(fragments),
        }

        self._prepare_and_start_frag_download(ctx, info_dict)

        extra_state = ctx.setdefault('extra_state', {
            'header_written': False,
            'mime_boundary': str(uuid.uuid4()).replace('-', ''),
        })

        frag_boundary = extra_state['mime_boundary']

        if not extra_state['header_written']:
            stub = self._gen_stub(
                fragments=fragments,
                frag_boundary=frag_boundary,
                title=title,
            )

            ctx['dest_stream'].write((
                'MIME-Version: 1.0\r\n'
                'From: <nowhere@yt-dlp.github.io.invalid>\r\n'
                'To: <nowhere@yt-dlp.github.io.invalid>\r\n'
                f'Subject: {self._escape_mime(title)}\r\n'
                'Content-type: multipart/related; '
                f'boundary="{frag_boundary}"; '
                'type="text/html"\r\n'
                f'X.yt-dlp.Origin: {origin}\r\n'
                '\r\n'
                f'--{frag_boundary}\r\n'
                'Content-Type: text/html; charset=utf-8\r\n'
                f'Content-Length: {len(stub)}\r\n'
                '\r\n'
                f'{stub}\r\n').encode())
            extra_state['header_written'] = True

        for i, fragment in enumerate(fragments):
            if (i + 1) <= ctx['fragment_index']:
                continue

            fragment_url = fragment.get('url')
            if not fragment_url:
                assert fragment_base_url
                fragment_url = urljoin(fragment_base_url, fragment['path'])

            success = self._download_fragment(ctx, fragment_url, info_dict)
            if not success:
                continue
            frag_content = self._read_fragment(ctx)

            frag_header = io.BytesIO()
            frag_header.write(
                b'--%b\r\n' % frag_boundary.encode('us-ascii'))
            frag_header.write(
                b'Content-ID: <%b>\r\n' % self._gen_cid(i, fragment, frag_boundary).encode('us-ascii'))
            frag_header.write(
                b'Content-type: %b\r\n' % f'image/{imghdr.what(h=frag_content) or "jpeg"}'.encode())
            frag_header.write(
                b'Content-length: %u\r\n' % len(frag_content))
            frag_header.write(
                b'Content-location: %b\r\n' % fragment_url.encode('us-ascii'))
            frag_header.write(
                b'X.yt-dlp.Duration: %f\r\n' % fragment['duration'])
            frag_header.write(b'\r\n')
            self._append_fragment(
                ctx, frag_header.getvalue() + frag_content + b'\r\n')

        ctx['dest_stream'].write(
            b'--%b--\r\n\r\n' % frag_boundary.encode('us-ascii'))
        return self._finish_frag_download(ctx, info_dict)



import re

from .common import InfoExtractor
from ..utils import (
    ExtractorError,
    determine_ext,
    float_or_none,
    int_or_none,
    js_to_json,
    traverse_obj,
    url_or_none,
)


class MuseAIIE(InfoExtractor):
    _VALID_URL = r'https?://(?:www\.)?muse\.ai/(?:v|embed)/(?P<id>\w+)'
    _TESTS = [{
        'url': 'https://muse.ai/embed/YdTWvUW',
        'md5': 'f994f9a38be1c3aaf9e37cbd7d76fe7c',
        'info_dict': {
            'id': 'YdTWvUW',
            'ext': 'mp4',
            'title': '2023-05-28-Grabien-1941111 (1)',
            'description': '',
            'uploader': 'Today News Africa',
            'uploader_id': 'TodayNewsAfrica',
            'upload_date': '20230528',
            'timestamp': 1685285044,
            'duration': 1291.3,
            'view_count': int,
            'availability': 'public',
        },
    }, {
        'url': 'https://muse.ai/v/gQ4gGAA-0756',
        'md5': '52dbfc78e865e56dc19a1715badc35e8',
        'info_dict': {
            'id': 'gQ4gGAA',
            'ext': 'mp4',
            'title': '0756',
            'description': 'md5:0ca1483f9aac423e9a96ad00bb3a0785',
            'uploader': 'Aerial.ie',
            'uploader_id': 'aerial',
            'upload_date': '20210306',
            'timestamp': 1615072842,
            'duration': 21.4,
            'view_count': int,
            'availability': 'public',
        },
    }]
    _WEBPAGE_TESTS = [{
        'url': 'https://muse.ai/docs',
        'playlist_mincount': 4,
        'info_dict': {
            'id': 'docs',
            'title': 'muse.ai | docs',
            'description': 'md5:6c0293431481582739c82ee8902687fa',
            'age_limit': 0,
            'thumbnail': 'https://muse.ai/static/imgs/poster-img-docs.png',
        },
        'params': {'allowed_extractors': ['all', '-html5']},
    }]
    _EMBED_REGEX = [r'<iframe[^>]*\bsrc=["\'](?P<url>https://muse\.ai/embed/\w+)']

    @classmethod
    def _extract_embed_urls(cls, url, webpage):
        yield from super()._extract_embed_urls(url, webpage)
        for embed_id in re.findall(r'<script>[^<]*\bMusePlayer\(\{[^}<]*\bvideo:\s*["\'](\w+)["\']', webpage):
            yield f'https://muse.ai/embed/{embed_id}'

    def _real_extract(self, url):
        video_id = self._match_id(url)
        webpage = self._download_webpage(f'https://muse.ai/embed/{video_id}', video_id)
        data = self._search_json(
            r'player\.setData\(', webpage, 'player data', video_id, transform_source=js_to_json)

        source_url = data['url']
        if not url_or_none(source_url):
            raise ExtractorError('Unable to extract video URL')

        formats = [{
            'url': source_url,
            'format_id': 'source',
            'quality': 1,
            **traverse_obj(data, {
                'ext': ('filename', {determine_ext}),
                'width': ('width', {int_or_none}),
                'height': ('height', {int_or_none}),
                'filesize': ('size', {int_or_none}),
            }),
        }]
        if source_url.endswith('/data'):
            base_url = f'{source_url[:-5]}/videos'
            formats.extend(self._extract_m3u8_formats(
                f'{base_url}/hls.m3u8', video_id, m3u8_id='hls', fatal=False))
            formats.extend(self._extract_mpd_formats(
                f'{base_url}/dash.mpd', video_id, mpd_id='dash', fatal=False))

        return {
            'id': video_id,
            'formats': formats,
            **traverse_obj(data, {
                'title': ('title', {str}),
                'description': ('description', {str}),
                'duration': ('duration', {float_or_none}),
                'timestamp': ('tcreated', {int_or_none}),
                'uploader': ('owner_name', {str}),
                'uploader_id': ('owner_username', {str}),
                'view_count': ('views', {int_or_none}),
                'age_limit': ('mature', {lambda x: 18 if x else None}),
                'availability': ('visibility', {lambda x: x if x in ('private', 'unlisted') else 'public'}),
            }),
        }



from .common import InfoExtractor
from ..utils import int_or_none, parse_duration, parse_iso8601


class NovaPlayIE(InfoExtractor):
    _VALID_URL = r'https?://play\.nova\.bg/video/[^?#]+/(?P<id>\d+)'
    _TESTS = [
        {
            'url': 'https://play.nova.bg/video/ochakvaite/season-0/ochakvaite-2022-07-22-sybudi-se-sat/606627',
            'md5': 'd79dff2d09d196c595a7290f48e33399',
            'info_dict': {
                'id': '606627',
                'ext': 'mp4',
                'title': 'Събуди се - събота по NOVA (23.07.2022)',
                'alt_title': 'ochakvaite/season-0/ochakvaite-2022-07-22-sybudi-se-sat',
                'duration': 29.0,
                'timestamp': 1658491547,
                'upload_date': '20220722',
                'thumbnail': 'https://nbg-img.fite.tv/img/606627_460x260.jpg',
                'description': '29 сек',
            },
        },
        {
            'url': 'https://play.nova.bg/video/ochakvaite/season-0/ochakvaite-2022-07-22-cherry-tazi/606609',
            'md5': 'f3e973e2ed1a5b9b3f498b1ab82d01b3',
            'info_dict': {
                'id': '606609',
                'ext': 'mp4',
                'title': 'Черешката на тортата - тази вечер по NOVA (22.07.2022)',
                'alt_title': 'ochakvaite/season-0/ochakvaite-2022-07-22-cherry-tazi',
                'duration': 29.0,
                'timestamp': 1658476303,
                'upload_date': '20220722',
                'thumbnail': 'https://nbg-img.fite.tv/img/606609_460x260.jpg',
                'description': '29 сек',
            },
        },
    ]

    _access_token = None

    def _real_extract(self, url):
        video_id = self._match_id(url)
        webpage = self._download_webpage(url, video_id)
        self._access_token = self._access_token or self._download_json(
            'https://play.nova.bg/api/client', None, note='Fetching access token')['accessToken']
        video_props = self._search_nextjs_data(webpage, video_id)['props']['pageProps']['video']
        m3u8_url = self._download_json(
            f'https://nbg-api.fite.tv/api/v2/videos/{video_id}/streams',
            video_id, headers={
                'x-flipps-user-agent': 'Flipps/75/9.7',
                'x-flipps-version': '2022-05-17',
                'Authorization': f'Bearer {self._access_token}',
            })[0]['links']['play']['href']
        formats = self._extract_m3u8_formats(m3u8_url, video_id, 'mp4', m3u8_id='hls')

        return {
            'id': video_id,
            'title': video_props['title'],
            'alt_title': video_props.get('slug'),
            'thumbnail': self._og_search_thumbnail(webpage),
            'description': self._og_search_description(webpage),
            'formats': formats,
            'duration': parse_duration(video_props['duration']),
            'timestamp': parse_iso8601(video_props['published_at']),
            'view_count': int_or_none(video_props['view_count']),
        }



from .common import InfoExtractor
from ..networking.exceptions import HTTPError
from ..utils import (
    ExtractorError,
    float_or_none,
    int_or_none,
    parse_resolution,
    str_or_none,
    traverse_obj,
    try_get,
    unified_timestamp,
    url_or_none,
    urljoin,
)


class PuhuTVIE(InfoExtractor):
    _VALID_URL = r'https?://(?:www\.)?puhutv\.com/(?P<id>[^/?#&]+)-izle'
    IE_NAME = 'puhutv'
    _TESTS = [{
        # film
        'url': 'https://puhutv.com/bi-kucuk-eylul-meselesi-izle',
        'md5': '4de98170ccb84c05779b1f046b3c86f8',
        'info_dict': {
            'id': '11909',
            'display_id': 'bi-kucuk-eylul-meselesi',
            'ext': 'mp4',
            'title': 'Bi Küçük Eylül Meselesi',
            'description': 'md5:c2ab964c6542b7b26acacca0773adce2',
            'thumbnail': r're:^https?://.*\.jpg$',
            'duration': 6176.96,
            'creator': 'Ay Yapım',
            'creators': ['Ay Yapım'],
            'timestamp': 1561062749,
            'upload_date': '20190620',
            'release_year': 2014,
            'view_count': int,
            'tags': list,
        },
    }, {
        # episode, geo restricted, bypassable with --geo-verification-proxy
        'url': 'https://puhutv.com/jet-sosyete-1-bolum-izle',
        'only_matching': True,
    }, {
        # 4k, with subtitles
        'url': 'https://puhutv.com/dip-1-bolum-izle',
        'only_matching': True,
    }]
    _SUBTITLE_LANGS = {
        'English': 'en',
        'Deutsch': 'de',
        'عربى': 'ar',
    }

    def _real_extract(self, url):
        display_id = self._match_id(url)

        info = self._download_json(
            urljoin(url, f'/api/slug/{display_id}-izle'),
            display_id)['data']

        video_id = str(info['id'])
        show = info.get('title') or {}
        title = info.get('name') or show['name']
        if info.get('display_name'):
            title = '{} {}'.format(title, info['display_name'])

        try:
            videos = self._download_json(
                f'https://puhutv.com/api/assets/{video_id}/videos',
                display_id, 'Downloading video JSON',
                headers=self.geo_verification_headers())
        except ExtractorError as e:
            if isinstance(e.cause, HTTPError) and e.cause.status == 403:
                self.raise_geo_restricted()
            raise

        urls = []
        formats = []

        for video in videos['data']['videos']:
            media_url = url_or_none(video.get('url'))
            if not media_url or media_url in urls:
                continue
            urls.append(media_url)

            playlist = video.get('is_playlist')
            if (video.get('stream_type') == 'hls' and playlist is True) or 'playlist.m3u8' in media_url:
                formats.extend(self._extract_m3u8_formats(
                    media_url, video_id, 'mp4', entry_protocol='m3u8_native',
                    m3u8_id='hls', fatal=False))
                continue

            quality = int_or_none(video.get('quality'))
            f = {
                'url': media_url,
                'ext': 'mp4',
                'height': quality,
            }
            video_format = video.get('video_format')
            is_hls = (video_format == 'hls' or '/hls/' in media_url or '/chunklist.m3u8' in media_url) and playlist is False
            if is_hls:
                format_id = 'hls'
                f['protocol'] = 'm3u8_native'
            elif video_format == 'mp4':
                format_id = 'http'
            else:
                continue
            if quality:
                format_id += f'-{quality}p'
            f['format_id'] = format_id
            formats.append(f)

        creator = try_get(
            show, lambda x: x['producer']['name'], str)

        content = info.get('content') or {}

        images = try_get(
            content, lambda x: x['images']['wide'], dict) or {}
        thumbnails = []
        for image_id, image_url in images.items():
            if not isinstance(image_url, str):
                continue
            if not image_url.startswith(('http', '//')):
                image_url = f'https://{image_url}'
            t = parse_resolution(image_id)
            t.update({
                'id': image_id,
                'url': image_url,
            })
            thumbnails.append(t)

        tags = []
        for genre in show.get('genres') or []:
            if not isinstance(genre, dict):
                continue
            genre_name = genre.get('name')
            if genre_name and isinstance(genre_name, str):
                tags.append(genre_name)

        subtitles = {}
        for subtitle in content.get('subtitles') or []:
            if not isinstance(subtitle, dict):
                continue
            lang = subtitle.get('language')
            sub_url = url_or_none(subtitle.get('url') or subtitle.get('file'))
            if not lang or not isinstance(lang, str) or not sub_url:
                continue
            subtitles[self._SUBTITLE_LANGS.get(lang, lang)] = [{
                'url': sub_url,
            }]

        return {
            'id': video_id,
            'display_id': display_id,
            'title': title,
            'description': info.get('description') or show.get('description'),
            'season_id': str_or_none(info.get('season_id')),
            'season_number': int_or_none(info.get('season_number')),
            'episode_number': int_or_none(info.get('episode_number')),
            'release_year': int_or_none(show.get('released_at')),
            'timestamp': unified_timestamp(info.get('created_at')),
            'creator': creator,
            'view_count': int_or_none(content.get('watch_count')),
            'duration': float_or_none(content.get('duration_in_ms'), 1000),
            'tags': tags,
            'subtitles': subtitles,
            'thumbnails': thumbnails,
            'formats': formats,
        }


class PuhuTVSerieIE(InfoExtractor):
    _VALID_URL = r'https?://(?:www\.)?puhutv\.com/(?P<id>[^/?#&]+)-detay'
    IE_NAME = 'puhutv:serie'
    _TESTS = [{
        'url': 'https://puhutv.com/deniz-yildizi-detay',
        'info_dict': {
            'title': 'Deniz Yıldızı',
            'id': 'deniz-yildizi',
        },
        'playlist_mincount': 205,
    }, {
        # a film detail page which is using same url with serie page
        'url': 'https://puhutv.com/bizim-icin-sampiyon-detay',
        'only_matching': True,
    }]

    def _extract_entries(self, seasons):
        for season in seasons:
            season_id = season.get('id')
            if not season_id:
                continue
            page = 1
            has_more = True
            while has_more is True:
                season = self._download_json(
                    f'https://appservice.puhutv.com/api/seasons/{season_id}/episodes?v=2',
                    season_id, f'Downloading page {page}', query={
                        'page': page,
                        'per': 100,
                    })['data']

                for episode in traverse_obj(season, ('episodes', lambda _, v: v.get('slug') or v['assets'][0]['slug'])):
                    slug = episode.get('slug') or episode['assets'][0]['slug']
                    yield self.url_result(
                        f'https://puhutv.com/{slug}', PuhuTVIE, episode.get('id'), episode.get('name'))

                page += 1
                has_more = traverse_obj(season, 'has_more')

    def _real_extract(self, url):
        playlist_id = self._match_id(url)

        info = self._download_json(
            urljoin(url, f'/api/slug/{playlist_id}-detay'),
            playlist_id)['data']

        seasons = info.get('seasons')
        if seasons:
            return self.playlist_result(
                self._extract_entries(seasons), playlist_id, info.get('name'))

        # For films, these are using same url with series
        video_id = info.get('slug')
        if video_id:
            video_id = video_id.removesuffix('-detay')
        else:
            video_id = info['assets'][0]['slug'].removesuffix('-izle')
        return self.url_result(
            f'https://puhutv.com/{video_id}-izle', PuhuTVIE, video_id)



import re

from .common import InfoExtractor
from ..networking import HEADRequest
from ..utils import (
    get_element_by_class,
    int_or_none,
    try_call,
    url_or_none,
    urlhandle_detect_ext,
)
from ..utils.traversal import traverse_obj


class Mx3BaseIE(InfoExtractor):
    _VALID_URL_TMPL = r'https?://(?:www\.)?%s/t/(?P<id>\w+)'
    _FORMATS = [{
        'url': 'player_asset',
        'format_id': 'default',
        'quality': 0,
    }, {
        'url': 'player_asset?quality=hd',
        'format_id': 'hd',
        'quality': 1,
    }, {
        'url': 'download',
        'format_id': 'download',
        'quality': 2,
    }, {
        'url': 'player_asset?quality=source',
        'format_id': 'source',
        'quality': 2,
    }]

    def _extract_formats(self, track_id):
        formats = []
        for fmt in self._FORMATS:
            format_url = f'https://{self._DOMAIN}/tracks/{track_id}/{fmt["url"]}'
            urlh = self._request_webpage(
                HEADRequest(format_url), track_id, fatal=False, expected_status=404,
                note=f'Checking for format {fmt["format_id"]}')
            if urlh and urlh.status == 200:
                formats.append({
                    **fmt,
                    'url': format_url,
                    'ext': urlhandle_detect_ext(urlh),
                    'filesize': int_or_none(urlh.headers.get('Content-Length')),
                })
        return formats

    def _real_extract(self, url):
        track_id = self._match_id(url)
        webpage = self._download_webpage(url, track_id)
        more_info = get_element_by_class('single-more-info', webpage)
        data = self._download_json(f'https://{self._DOMAIN}/t/{track_id}.json', track_id, fatal=False)

        def get_info_field(name):
            return self._html_search_regex(
                rf'<dt[^>]*>\s*{name}\s*</dt>\s*<dd[^>]*>(.*?)</dd>',
                more_info, name, default=None, flags=re.DOTALL)

        return {
            'id': track_id,
            'formats': self._extract_formats(track_id),
            'genre': self._html_search_regex(
                r'<div\b[^>]+class="single-band-genre"[^>]*>([^<]+)</div>', webpage, 'genre', default=None),
            'release_year': int_or_none(get_info_field('Year of creation')),
            'description': get_info_field('Description'),
            'tags': try_call(lambda: get_info_field('Tag').split(', '), list),
            **traverse_obj(data, {
                'title': ('title', {str}),
                'artist': (('performer_name', 'artist'), {str}),
                'album_artist': ('artist', {str}),
                'composer': ('composer_name', {str}),
                'thumbnail': (('picture_url_xlarge', 'picture_url'), {url_or_none}),
            }, get_all=False),
        }


class Mx3IE(Mx3BaseIE):
    _DOMAIN = 'mx3.ch'
    _VALID_URL = Mx3BaseIE._VALID_URL_TMPL % re.escape(_DOMAIN)
    _TESTS = [{
        'url': 'https://mx3.ch/t/1Cru',
        'md5': '7ba09e9826b4447d4e1ce9d69e0e295f',
        'info_dict': {
            'id': '1Cru',
            'ext': 'wav',
            'artist': 'Godina',
            'album_artist': 'Tortue Tortue',
            'composer': 'Olivier Godinat',
            'genre': 'Rock',
            'thumbnail': 'https://mx3.ch/pictures/mx3/file/0101/4643/square_xlarge/1-s-envoler-1.jpg?1630272813',
            'title': "S'envoler",
            'release_year': 2021,
            'tags': [],
        },
    }, {
        'url': 'https://mx3.ch/t/1LIY',
        'md5': '48293cb908342547827f963a5a2e9118',
        'info_dict': {
            'id': '1LIY',
            'ext': 'mov',
            'artist': 'Tania Kimfumu',
            'album_artist': 'The Broots',
            'composer': 'Emmanuel Diserens',
            'genre': 'Electro',
            'thumbnail': 'https://mx3.ch/pictures/mx3/file/0110/0003/video_xlarge/frame_0000.png?1686963670',
            'title': 'The Broots-Larytta remix "Begging For Help"',
            'release_year': 2023,
            'tags': ['the broots', 'cassata records', 'larytta'],
            'description': '"Begging for Help" Larytta Remix Official Video\nRealized By Kali Donkilie in 2023',
        },
    }, {
        'url': 'https://mx3.ch/t/1C6E',
        'md5': '1afcd578493ddb8e5008e94bb6d97e25',
        'info_dict': {
            'id': '1C6E',
            'ext': 'wav',
            'artist': 'Alien Bubblegum',
            'album_artist': 'Alien Bubblegum',
            'composer': 'Alien Bubblegum',
            'genre': 'Punk',
            'thumbnail': 'https://mx3.ch/pictures/mx3/file/0101/1551/square_xlarge/pandora-s-box-cover-with-title.png?1627054733',
            'title': 'Wide Awake',
            'release_year': 2021,
            'tags': ['alien bubblegum', 'bubblegum', 'alien', 'pop punk', 'poppunk'],
        },
    }]


class Mx3NeoIE(Mx3BaseIE):
    _DOMAIN = 'neo.mx3.ch'
    _VALID_URL = Mx3BaseIE._VALID_URL_TMPL % re.escape(_DOMAIN)
    _TESTS = [{
        'url': 'https://neo.mx3.ch/t/1hpd',
        'md5': '6d9986bbae5cac3296ec8813bf965eb2',
        'info_dict': {
            'id': '1hpd',
            'ext': 'wav',
            'artist': 'Baptiste Lopez',
            'album_artist': 'Kammerorchester Basel',
            'composer': 'Jannik Giger',
            'genre': 'Composition, Orchestra',
            'title': 'Troisième œil. Für Kammerorchester (2023)',
            'thumbnail': 'https://neo.mx3.ch/pictures/neo/file/0000/0241/square_xlarge/kammerorchester-basel-group-photo-2_c_-lukasz-rajchert.jpg?1560341252',
            'release_year': 2023,
            'tags': [],
        },
    }]


class Mx3VolksmusikIE(Mx3BaseIE):
    _DOMAIN = 'volksmusik.mx3.ch'
    _VALID_URL = Mx3BaseIE._VALID_URL_TMPL % re.escape(_DOMAIN)
    _TESTS = [{
        'url': 'https://volksmusik.mx3.ch/t/Zx',
        'md5': 'dd967a7b0c1ef898f3e072cf9c2eae3c',
        'info_dict': {
            'id': 'Zx',
            'ext': 'mp3',
            'artist': 'Ländlerkapelle GrischArt',
            'album_artist': 'Ländlerkapelle GrischArt',
            'composer': 'Urs Glauser',
            'genre': 'Instrumental, Graubünden',
            'title': 'Chämilouf',
            'thumbnail': 'https://volksmusik.mx3.ch/pictures/vxm/file/0000/3815/square_xlarge/grischart1.jpg?1450530120',
            'release_year': 2012,
            'tags': [],
        },
    }]



from .bunnycdn import BunnyCdnIE
from .common import InfoExtractor
from ..utils import make_archive_id, try_get, unified_timestamp


class SovietsClosetBaseIE(InfoExtractor):
    MEDIADELIVERY_REFERER = {'Referer': 'https://iframe.mediadelivery.net/'}

    def parse_nuxt_jsonp(self, nuxt_jsonp_url, video_id, name):
        nuxt_jsonp = self._download_webpage(nuxt_jsonp_url, video_id, note=f'Downloading {name} __NUXT_JSONP__')
        return self._search_nuxt_data(nuxt_jsonp, video_id, '__NUXT_JSONP__')

    def video_meta(self, video_id, game_name, category_name, episode_number, stream_date):
        title = game_name
        if category_name and category_name != 'Misc':
            title += f' - {category_name}'
        if episode_number:
            title += f' #{episode_number}'

        timestamp = unified_timestamp(stream_date)

        return {
            'id': video_id,
            'title': title,
            'http_headers': self.MEDIADELIVERY_REFERER,
            'uploader': 'SovietWomble',
            'creator': 'SovietWomble',
            'release_timestamp': timestamp,
            'timestamp': timestamp,
            'uploader_id': 'SovietWomble',
            'uploader_url': 'https://www.twitch.tv/SovietWomble',
            'was_live': True,
            'availability': 'public',
            'series': game_name,
            'season': category_name,
            'episode_number': episode_number,
        }


class SovietsClosetIE(SovietsClosetBaseIE):
    _VALID_URL = r'https?://(?:www\.)?sovietscloset\.com/video/(?P<id>[0-9]+)/?'
    _TESTS = [
        {
            'url': 'https://sovietscloset.com/video/1337',
            'md5': 'bd012b04b261725510ca5383074cdd55',
            'info_dict': {
                'id': '2f0cfbf4-3588-43a9-a7d6-7c9ea3755e67',
                'ext': 'mp4',
                'title': 'The Witcher #13',
                'thumbnail': r're:^https?://.*\.b-cdn\.net/2f0cfbf4-3588-43a9-a7d6-7c9ea3755e67/thumbnail\.jpg$',
                'uploader': 'SovietWomble',
                'creator': 'SovietWomble',
                'release_timestamp': 1492091580,
                'release_date': '20170413',
                'timestamp': 1492091580,
                'upload_date': '20170413',
                'uploader_id': 'SovietWomble',
                'uploader_url': 'https://www.twitch.tv/SovietWomble',
                'duration': 7008,
                'was_live': True,
                'availability': 'public',
                'series': 'The Witcher',
                'season': 'Misc',
                'episode_number': 13,
                'episode': 'Episode 13',
                'creators': ['SovietWomble'],
                'description': '',
                '_old_archive_ids': ['sovietscloset 1337'],
            },
        },
        {
            'url': 'https://sovietscloset.com/video/1105',
            'md5': '89fa928f183893cb65a0b7be846d8a90',
            'info_dict': {
                'id': 'c0e5e76f-3a93-40b4-bf01-12343c2eec5d',
                'ext': 'mp4',
                'title': 'Arma 3 - Zeus Games #5',
                'uploader': 'SovietWomble',
                'thumbnail': r're:^https?://.*\.b-cdn\.net/c0e5e76f-3a93-40b4-bf01-12343c2eec5d/thumbnail\.jpg$',
                'creator': 'SovietWomble',
                'release_timestamp': 1461157200,
                'release_date': '20160420',
                'timestamp': 1461157200,
                'upload_date': '20160420',
                'uploader_id': 'SovietWomble',
                'uploader_url': 'https://www.twitch.tv/SovietWomble',
                'duration': 8805,
                'was_live': True,
                'availability': 'public',
                'series': 'Arma 3',
                'season': 'Zeus Games',
                'episode_number': 5,
                'episode': 'Episode 5',
                'creators': ['SovietWomble'],
                'description': '',
                '_old_archive_ids': ['sovietscloset 1105'],
            },
        },
    ]

    def _real_extract(self, url):
        video_id = self._match_id(url)
        webpage = self._download_webpage(url, video_id)

        static_assets_base = self._search_regex(r'(/_nuxt/static/\d+)', webpage, 'staticAssetsBase')
        static_assets_base = f'https://sovietscloset.com{static_assets_base}'

        stream = self.parse_nuxt_jsonp(f'{static_assets_base}/video/{video_id}/payload.js', video_id, 'video')['stream']

        return self.url_result(
            f'https://iframe.mediadelivery.net/embed/5105/{stream["bunnyId"]}', ie=BunnyCdnIE, url_transparent=True,
            **self.video_meta(
                video_id=video_id, game_name=stream['game']['name'],
                category_name=try_get(stream, lambda x: x['subcategory']['name'], str),
                episode_number=stream.get('number'), stream_date=stream.get('date')),
            _old_archive_ids=[make_archive_id(self, video_id)])


class SovietsClosetPlaylistIE(SovietsClosetBaseIE):
    _VALID_URL = r'https?://(?:www\.)?sovietscloset\.com/(?!video)(?P<id>[^#?]+)'
    _TESTS = [

        {
            'url': 'https://sovietscloset.com/The-Witcher',
            'info_dict': {
                'id': 'The-Witcher',
                'title': 'The Witcher',
            },
            'playlist_mincount': 31,
        },
        {
            'url': 'https://sovietscloset.com/Arma-3/Zeus-Games',
            'info_dict': {
                'id': 'Arma-3/Zeus-Games',
                'title': 'Arma 3 - Zeus Games',
            },
            'playlist_mincount': 3,
        },
        {
            'url': 'https://sovietscloset.com/arma-3/zeus-games/',
            'info_dict': {
                'id': 'arma-3/zeus-games',
                'title': 'Arma 3 - Zeus Games',
            },
            'playlist_mincount': 3,
        },
        {
            'url': 'https://sovietscloset.com/Total-War-Warhammer',
            'info_dict': {
                'id': 'Total-War-Warhammer',
                'title': 'Total War: Warhammer - Greenskins',
            },
            'playlist_mincount': 33,
        },
    ]

    def _real_extract(self, url):
        playlist_id = self._match_id(url)
        if playlist_id.endswith('/'):
            playlist_id = playlist_id[:-1]

        webpage = self._download_webpage(url, playlist_id)

        static_assets_base = self._search_regex(r'(/_nuxt/static/\d+)', webpage, 'staticAssetsBase')
        static_assets_base = f'https://sovietscloset.com{static_assets_base}'

        sovietscloset = self.parse_nuxt_jsonp(f'{static_assets_base}/payload.js', playlist_id, 'global')['games']

        if '/' in playlist_id:
            game_slug, category_slug = playlist_id.lower().split('/')
        else:
            game_slug = playlist_id.lower()
            category_slug = 'misc'

        game = next(game for game in sovietscloset if game['slug'].lower() == game_slug)
        category = next((cat for cat in game['subcategories'] if cat.get('slug', '').lower() == category_slug),
                        game['subcategories'][0])
        category_slug = category.get('slug', '').lower() or category_slug
        playlist_title = game.get('name') or game_slug
        if category_slug != 'misc':
            playlist_title += f' - {category.get("name") or category_slug}'
        entries = [{
            **self.url_result(f'https://sovietscloset.com/video/{stream["id"]}', ie=SovietsClosetIE.ie_key()),
            **self.video_meta(
                video_id=stream['id'], game_name=game['name'], category_name=category.get('name'),
                episode_number=i + 1, stream_date=stream.get('date')),
        } for i, stream in enumerate(category['streams'])]

        return self.playlist_result(entries, playlist_id, playlist_title)



import json

from .common import InfoExtractor
from ..utils import int_or_none


class PodomaticIE(InfoExtractor):
    _WORKING = False
    IE_NAME = 'podomatic'
    _VALID_URL = r'''(?x)
                    (?P<proto>https?)://
                        (?:
                            (?P<channel>[^.]+)\.podomatic\.com/entry|
                            (?:www\.)?podomatic\.com/podcasts/(?P<channel_2>[^/]+)/episodes
                        )/
                        (?P<id>[^/?#&]+)
                '''

    _TESTS = [{
        'url': 'http://scienceteachingtips.podomatic.com/entry/2009-01-02T16_03_35-08_00',
        'md5': '84bb855fcf3429e6bf72460e1eed782d',
        'info_dict': {
            'id': '2009-01-02T16_03_35-08_00',
            'ext': 'mp3',
            'uploader': 'Science Teaching Tips',
            'uploader_id': 'scienceteachingtips',
            'title': '64.  When the Moon Hits Your Eye',
            'duration': 446,
        },
    }, {
        'url': 'http://ostbahnhof.podomatic.com/entry/2013-11-15T16_31_21-08_00',
        'md5': 'd2cf443931b6148e27638650e2638297',
        'info_dict': {
            'id': '2013-11-15T16_31_21-08_00',
            'ext': 'mp3',
            'uploader': 'Ostbahnhof / Techno Mix',
            'uploader_id': 'ostbahnhof',
            'title': 'Einunddreizig',
            'duration': 3799,
        },
    }, {
        'url': 'https://www.podomatic.com/podcasts/scienceteachingtips/episodes/2009-01-02T16_03_35-08_00',
        'only_matching': True,
    }]

    def _real_extract(self, url):
        mobj = self._match_valid_url(url)
        video_id = mobj.group('id')
        channel = mobj.group('channel') or mobj.group('channel_2')

        json_url = ('{}://{}.podomatic.com/entry/embed_params/{}?permalink=true&rtmp=0'.format(
            mobj.group('proto'), channel, video_id))
        data_json = self._download_webpage(
            json_url, video_id, 'Downloading video info')
        data = json.loads(data_json)

        video_url = data['downloadLink']
        if not video_url:
            video_url = '{}/{}'.format(data['streamer'].replace('rtmp', 'http'), data['mediaLocation'])
        uploader = data['podcast']
        title = data['title']
        thumbnail = data['imageLocation']
        duration = int_or_none(data.get('length'), 1000)

        return {
            'id': video_id,
            'url': video_url,
            'title': title,
            'uploader': uploader,
            'uploader_id': channel,
            'thumbnail': thumbnail,
            'duration': duration,
        }
